package ru.ispras.telecomtesk.tlsv13.TLSserverSpecification;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.TLSSocket.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.TLSContentType;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.LoginInfo;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

public
class TLSServerScenProp
{

    public
    String targetServer;

    public
    int targetPort;

    public
    String socketTimeout;

    public
    LoginInfo scenProp;

    public
    TLSConnection current_connection = null;

    public
    TLSServerSpec target = new TLSServerSpec(  );

    public
    InetAddress hostIP;

    public
    String applicationDataString = null;

    public
    long connectCounter = 0;

    public
    long scenarioCounter = 0;

    public
    PrintWriter logFile = null;

    public
    TLSServerScenProp( String logFileName
    ,String currentClassName)
    {
        super( );
        scenProp = new LoginInfo(  );
        scenProp.readFromFile( logFileName );
        try
        {
            logFile = new PrintWriter( new OutputStreamWriter( new FileOutputStream( currentClassName + "_errorlog.txt", false ) ) );
            DebugInfo.printMessage( 1, "open log file " + currentClassName + "_errorlog.txt" );
        }
        catch( FileNotFoundException err )
        {
            DebugInfo.printMessage( 1, "open log file Exception: " + err );
            throw new IllegalArgumentException( "open log file Exception" );
        }
        targetServer = scenProp.getExistingValue( "server" );
        if( targetServer == null )
        {
            throw new IllegalArgumentException( "Server not specified" );
        }

        String targetPortString = scenProp.getExistingValue( "portTLS" );
        if( targetPortString == null )
        {
            throw new IllegalArgumentException( "Server port not specified" );
        }
        targetPort = Integer.parseInt( targetPortString );
        applicationDataString = scenProp.getValue( "applicationData", "Test message" );
        socketTimeout = scenProp.getValue( "timeout" );
        try
        {
            hostIP = InetAddress.getLocalHost(  );
        }
        catch( UnknownHostException err )
        {
            throw new RuntimeException( "Failed to initialize scenario", err );
        }
        target.sock = new TLSSocket( targetServer, targetPort, Integer.parseInt( socketTimeout ) );
    }

    public
    Exception msgSetTestContent( RecordLayer rlmsg
    ,byte[] aad
    ,Object msg
    ,String scenName)
    {

        Exception res = null;
        current_connection.setAAD( aad );
        try
        {
            rlmsg.setContent( msg, null, current_connection );
            DebugInfo.printDump( 5, rlmsg.toByteArray(  ), scenName );
        }
        catch( Exception err )
        {
            DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
            DebugInfo.logtrace( logFile, "===Error! scenario setContent Exception: " + err );
            target.disconnect(  );
            res = err;
        }
        current_connection.setAAD( null );
        return res;
    }


    public
    Exception msgSetLargeContentAndSend( Object msg
    ,byte[] padding
    ,String scenName)
    {
        return msgSetLargeContentAndSend( new RecordLayer( msg ), msg, padding, scenName );
    }


    public
    Exception msgSetContentAndSend( Object msg
    ,byte[] padding
    ,String scenName)
    {
        return msgSetContentAndSend( null, new RecordLayer( msg ), msg, padding, scenName );
    }


    public
    Exception msgSetContentAndSend( TLSContentType tp
    ,Object msg
    ,byte[] padding
    ,String scenName)
    {
        return msgSetContentAndSend( tp, new RecordLayer( msg ), msg, padding, scenName );
    }


    public
    Exception msgSetContentAndSend( RecordLayer rlmsg
    ,Object msg
    ,byte[] padding
    ,String scenName)
    {
        return msgSetContentAndSend( null, rlmsg, msg, padding, scenName );
    }


    public
    Exception msgSetContentAndSend( TLSContentType tp
    ,RecordLayer rlmsg
    ,Object msg
    ,byte[] padding
    ,String scenName)
    {

        Exception res = null;
        try
        {
            if( tp == null )
                rlmsg.setContent( msg, padding, current_connection );

            else
                rlmsg.setContent( msg, padding, current_connection, tp );

            DebugInfo.printDump( 5, rlmsg.toByteArray(  ), scenName );
        }
        catch( Exception err )
        {
            DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
            DebugInfo.logtrace( logFile, "===Error! scenario setContent Exception: " + err );
            target.disconnect(  );
            return res;
        }
        res = msgSend( rlmsg, scenName );
        return res;
    }


    public
    Exception msgSetLargeContentAndSend( RecordLayer rlmsg
    ,Object msg
    ,byte[] padding
    ,String scenName)
    {

        Exception res = null;
        try
        {
            rlmsg.setLargeContent( msg, padding, current_connection );
            DebugInfo.printDump( 5, rlmsg.toByteArray(  ), scenName );
        }
        catch( Exception err )
        {
            DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
            DebugInfo.logtrace( logFile, "===Error! scenario setLargeContent Exception: " + err );
            target.disconnect(  );
            return res;
        }
        res = msgSend( rlmsg, scenName );
        return res;
    }


    public
    Exception msgSend( RecordLayer rlmsg
    ,String scenName)
    {
        return msgSend( rlmsg, scenName, true );
    }


    public
    Exception msgSendWithoutAlert( RecordLayer rlmsg
    ,String scenName)
    {
        return msgSend( rlmsg, scenName, false );
    }


    public
    Exception msgSend( RecordLayer rlmsg
    ,String scenName
    ,boolean sendAlert)
    {

        Exception res = null;
        try
        {

            ArrayList<RecordLayer> response = target.send( rlmsg, current_connection );
            if( null == response )
            {
                DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
                DebugInfo.logtrace( logFile, "===Error! scenario send response null, may be decrypt failed" );
                target.disconnectWithSendAlert = 80;
                target.DisconnectWithAlert( current_connection );
            }
            else
                if( (target.disconnectWithSendAlert != -1) && (sendAlert) )
                {
                    target.DisconnectWithAlert( current_connection );
                }

        }
        catch( jatva.lang.TestException terr )
        {
            DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
            DebugInfo.logtrace( logFile, "===Error! scenario send TestException: " + terr );
            target.DisconnectWithAlert( current_connection );
            res = new RuntimeException( "scenario" + scenName, terr );
        }
        catch( Exception err )
        {
            DebugInfo.logtrace( logFile, "===Error! counter: " + scenarioCounter + " scenario" + scenName );
            DebugInfo.logtrace( logFile, "===Error! scenario send Exception: " + err );
            target.disconnect(  );
            res = err;
        }
        return res;
    }

}
