package ru.ispras.telecomtesk.tlsv13;

public class ConvertTypes {

	static public byte[] hton(long x) {
		byte[] res = new byte[8];
		for (int i = 0; i < 8; i++)
		{
			res[i] = (new Long(x >>> 56)).byteValue();
			x <<= 8;
		}
		return res;
	}

	static public byte[] hton(int x) {
		byte[] res = new byte[4];
		for (int i = 0; i < 4; i++)
		{
			res[i] = (new Integer(x >>> 24)).byteValue();
			x <<= 8;
		}
		return res;
	}

	static public byte[] hton(short x) {
		byte[] res = new byte[2];
		res[0] = (byte)(x >>> 8);
		res[1] = (byte)x;
		return res;
	}

	// byte array up to 4 bytes long will be converted
	static public int ntoh(byte[] x)
	{
		int res = 0, i = 0;
		for (; i < 4 && i < x.length; i++)
		{
			res <<= 8;
			res |= (x[i] & 0xFF);
		}
		return res;
	}

	static public int ntoh(byte[] x, int pos)
	{
		int res = 0, i = 0;
		for (; i < 4 && i+pos < x.length; i++)
		{
			res <<= 8;
			res |= (x[i+pos] & 0xFF);
		}
		return res;
	}

	// byte array up to len bytes long will be converted beginning from position pos
	static public int ntoh(byte[] x, int pos, int len) {
    	int res = 0;
    	while(len-- > 0) {
    		res <<= 8;
    		res |= (x[pos++] & 0xFF);
    	}
    	return res;
    }

	// byte array up to 2 bytes long will be converted
	static public short ntohs(byte[] x)
	{
		short res = 0;
		if(x.length >= 2) {
				res = (short)(x[0] & 0xFF);
				res <<= 8;
				res |= (x[1] & 0xFF);
		}
		return res;
	}

	static public short ntohs(byte[] x, int pos)
	{
		short res = 0;
		if((pos >= 0)&&(x.length >= pos + 2)) {
				res = (short)(x[pos++] & 0xFF);
				res <<= 8;
				res |= (x[pos] & 0xFF);
		}
		return res;
	}

	static public int int24bits(int a)
	{
		a &= 0x00ffffff;
		return a;
	}

	static public int unsignedByteToInt(byte b)
	{
		return (int) b & 0xFF;
	}

	public static String codeToString(int code, String name) {
		String tmpstr = String.format("%04x", code);
		if(tmpstr.length() > 4)
			tmpstr = tmpstr.substring(tmpstr.length()-4);
		return "0x" + tmpstr + " [" + name + "]";
	}

	public static byte[] combineArrays(byte[][] data) {
		int curpos = 0, datalen = 0;
		byte[] dataFull = null;
		
		if(data != null) {
			for(byte[] dataCur : data) {
				if(dataCur != null)
					datalen += dataCur.length;
			}
			dataFull = new byte[datalen];
			for(byte[] dataCur : data) {
				if(dataCur != null) {
					System.arraycopy(dataCur, 0, dataFull, curpos, dataCur.length);
					curpos += dataCur.length;
				}
			}
		}
		return dataFull;
	}

	public static boolean compareArrays(byte[] ar1, byte[] ar2) {
		boolean res = false;
		int i;
		
		if((ar1 != null)&&(ar2 != null)&&(ar1.length == ar2.length)) {
			for(i=0; i < ar1.length; i++) {
				if(ar1[i] != ar2[i])
					break;
			}
			if(i == ar1.length)
				res = true;
		}
		return res;
	}

}