package ru.ispras.telecomtesk.tlsv13;

import java.io.PrintWriter;

public class DebugInfo {
	public static int debugLevel = 1;
	public static int maxDumpSize = 200;
	
	public static void printMessage(int level, String str) {
		if(debugLevel >= level)
			System.out.println(str);
	}

	public static void printMessageWithSeparator(int level, String str) {
		if(debugLevel >= level) {
			System.out.println("\n========================\n");
			printMessage(level,str);
		}
	}

	public static void printDump(int level, byte[] array, String message) {
		printDump(level,array,0,message);
	}
	
	public static void printDump(int level, byte[] array, int offset, String message) {

		if(debugLevel < level)	return;
		
	    if(message != null)
	    	System.out.println(message);

	    if(array == null) {
	    	System.out.println("data NULL pointer");
	    	return;
	    }

	    if(array.length - offset <= 0) {
	    	System.out.println("no data");
	    	return;
	    }

	    int i,k;
	    int j = 0;

	    if(array.length - offset > maxDumpSize) {
	    	System.out.println("DebugInfo::printDump: Array is too big: " + (array.length - offset));
	    	return;
	    }
	    
	    for (i=offset,k = 1; i < array.length; i++, k++) {
	        if( i == array.length-1 || k == 16) {
	        	System.out.printf("%02x\n", (~(~0 << 8) & array[i]));
	            j = 0;
	            k = 0;
	        }
	        else if (j != 0) {
	        	System.out.printf("%02x ", (~(~0 << 8) & array[i]));
	            j--;
	        }
	        else {
	        	System.out.printf("%02x", (~(~0 << 8) & array[i]));
	            j++;
	        }
	    }
	}

	public static void printDumpWithLength(int level, byte[] array, String message) {
		if(array != null) {
			printDump(level, array, message + " [" + array.length + " bytes] ");
		} else {
			printDump(level, array, message);
		}
	}
	
	public static void printDumpWithLength(int level, byte[] array, int offset, String message) {
		if(array != null) {
			printDump(level, array, offset, message + " [" + (array.length - offset) + " bytes] ");
		} else {
			printDump(level, array, offset, message);
		}
	}
	
	public static void printDump(int level, byte[] array, int offset, int length, String message) {

		if(debugLevel < level)
			return;
		
	    if(message != null)
	    	System.out.println(message);

	    if(array == null || offset > array.length) {
	    	System.out.println("data NULL pointer");
	    	return;
	    }

	    if(array.length == 0) {
	    	System.out.println("no data");
	    	return;
	    }

		int i;
	    int j = 0, k = 1, len;
	    
	    len = array.length >= length + offset ? length + offset : array.length;

	    if(len > maxDumpSize) {
	    	System.out.println("DebugInfo::printDump: Array is too big: " + len);
	    	return;
	    }
	    
	    for (i=offset; i < len; i++, k++) {
	        if( i == len-1 || k == 16) {
	        	System.out.printf("%02x\n", (~(~0 << 8) & array[i]));
	            j = 0;
	            k = 0;
	        }
	        else if (j != 0) {
	        	System.out.printf("%02x ", (~(~0 << 8) & array[i]));
	            j--;
	        }
	        else {
	        	System.out.printf("%02x", (~(~0 << 8) & array[i]));
	            j++;
	        }
	    }
	}

	public static void printDump(int level, byte[][] array, String message) {

		if(debugLevel < level)	return;
		
	    if(message != null)
	    	System.out.println(message);

	    if(array == null) {
	    	System.out.println("data NULL pointer");
	    	return;
	    }

	    for (byte[] ar : array) {
    		printDump(level,ar,"");
	    }
	}

	public static byte[] hexStringToByteArray(String s) {
		int len = s.length();
		byte[] data = new byte[len / 2];
		for (int i = 0; i < len; i += 2) {
		    data[i / 2] = (byte) ((Character.digit(s.charAt(i), 16) << 4) + Character
			    .digit(s.charAt(i + 1), 16));
		}
		return data;
	}

	public static void printDumpToFile(PrintWriter logFile, byte[] array, String message) {
	    if(message != null)
	    	logFile.println(message);

	    if(array == null) {
	    	logFile.println("data NULL pointer");
	    	return;
	    }

	    if(array.length == 0) {
	    	logFile.println("no data");
	    	return;
	    }

	    int i;
	    int j = 0, k = 1, len = array.length;

	    if(len > maxDumpSize) {
	    	logFile.println("DebugInfo::printDump: Array is too big: " + len);
	    	return;
	    }
	    
	    for (i=0; i < len; i++, k++) {
	        if( i == len-1 || k == 16) {
	        	logFile.printf("%02x\n", (~(~0 << 8) & array[i]));
	            j = 0;
	            k = 0;
	        }
	        else if (j != 0) {
	        	logFile.printf("%02x ", (~(~0 << 8) & array[i]));
	            j--;
	        }
	        else {
	        	logFile.printf("%02x", (~(~0 << 8) & array[i]));
	            j++;
	        }
	    }
	}

	public static void logtrace (PrintWriter logFile, String msg) {
		logFile.println(msg);
		logFile.flush();
		DebugInfo.printMessage(1,msg);
	}

	public static void logtrace (PrintWriter logFile, byte[] data, String msg) {
		printDumpToFile(logFile, data, msg);
		logFile.flush();
		DebugInfo.printDump(1, data, msg);
	}
}