package ru.ispras.telecomtesk.tlsv13.HandshakeMessages;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.TLSExtension;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.TLSExtensionsList;

import java.util.ArrayList;

public class CertificateEntry {    // HandshakeMessages Certificate
	private byte[] data;             // <1..2^24-1>
	private ArrayList<TLSExtension> extensions; // <0..2^16-1> 

	public CertificateEntry(byte[] data) {
		this(data, null);
	}

	public CertificateEntry(byte[] data, ArrayList<TLSExtension> extensions) {
		this.data = data != null ? data : new byte[0];
		this.extensions = extensions != null ? extensions : new ArrayList<TLSExtension>();
	}

	public int getSize() {
		return (3 + data.length + 2 + TLSExtensionsList.getExtensionLength(extensions));
	}

	public byte[] getData() {
		return data;
	}

	public ArrayList<TLSExtension> getExtensions() {
		return extensions;
	}

	public byte[] toByteArray() {
		// create byte array
		int len = getSize();
		byte[] ba = new byte[len];
		// write 3 bytes of certificate length and certificate
		byte[] ba_len = ConvertTypes.hton(data.length); // return byte[4]
		System.arraycopy(ba_len, 1, ba, 0, 3);
		System.arraycopy(data, 0, ba, 3, data.length);
		//write extensions
		TLSExtensionsList.toByteArray(ba, (3 + data.length), extensions);
		return ba;
	}

	public void print() {
		DebugInfo.printMessage(1,"CertificateEntry: certificate length " + data.length);
		TLSExtensionsList.print(this.extensions);
	}

    public static CertificateEntry parseByteArray(byte[] msg, int pos) {
    	if(msg.length - pos < 3) {
    		DebugInfo.printMessage(4,"===Error! CertificateEntry::parseByteArray: certificate length failed");
    		return null;
		}
    	int certlen = ConvertTypes.ntoh(msg,pos,3);
    	pos += 3;
    	
    	if(msg.length - pos < certlen) {
    		DebugInfo.printMessage(4,"===Error! CertificateEntry::parseByteArray: data length wrong");
    		return null;
		}
    	byte[] data = new byte[certlen];
		System.arraycopy(msg, pos, data, 0, certlen);
    	pos += certlen;
    	
    	// parse Extensions
    	ArrayList<TLSExtension> extensions = TLSExtensionsList.parseTLSExtensionsList(msg, pos);
    	
    	
    	return new CertificateEntry(data, extensions);
    }
    	
}
