package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

public class ServerNameExt extends TLSExtension
{
	public ArrayList<ServerName> data; // <0..2^16-1> bytes 

	public ServerNameExt() {
		this.type = 0;
		this.data = new ArrayList<ServerName>();
	}

	public ServerNameExt(ArrayList<ServerName> data) {
		this.type = 0;
		this.data = data;
	}

	public ServerNameExt(ServerName data) {
		this();
		this.data.add(data);
	}

	public ServerNameExt(String data) {
		this(new ServerName(data));
	}

	public ServerNameExt(int length , ArrayList<ServerName> data) {
		this(data);
		this.setLength(length);
	}

	public void addServerName(ServerName name) {
		this.data.add(name);
	}

	public void addServerName(String name) {
		this.data.add(new ServerName(name));
	}

	@Override
	public int getLengthFieldSize() {
		return 2;
	}

	@Override
	public int getDataSize() {
		int res = 0;  
		if(data != null) {
			for (int j=0; j < data.size(); j++) {
				res += (data.get(j)).getSize();
			}
		}
		return res;
	}

	@Override
	public void toByteArrayData(byte[] msg, int position) {
		if(data != null) {
			for (int j=0; j < data.size(); j++) {
                int curPos = (data.get(j)).toByteArray(msg,position);
				if(curPos >= 0) 
					position = curPos;
				else 
					DebugInfo.printMessage(1,"!!! Error: ServerNameExt toByteArray return null, [" + j + "]");
			}
		}
	}

	@Override
	public void printData() {
		DebugInfo.printMessage(1,"  ServerNameExt");
		if(data != null) {
			for(int j=0; j<data.size(); j++)
				(data.get(j)).print();
		}
	}

	public static TLSExtension parseByteArray(int lengthExt, byte[] msg, int pos) {
		int len;
		int lengthFieldSize = 2;
    	if(lengthExt < lengthFieldSize) {
    		DebugInfo.printMessage(4,"===Error! ServerNameExt::parseByteArray: Extension length failed");
    		return null;
    	}
		len = ConvertTypes.ntohs(msg,pos);
		pos += lengthFieldSize;

    	if(lengthExt != lengthFieldSize + len) {
    		DebugInfo.printMessage(4,"===Error! ServerNameExt::parseByteArray: Extension length wrong");
    		return null;
    	}
    	
    	ArrayList<ServerName> extdata = new ArrayList<ServerName>();
    	int keylen;

		while(len > 0) {
			ServerName curKeyEntry = ServerName.parseByteArray(msg,pos);
			if(curKeyEntry == null) 
				return null;
    		extdata.add(curKeyEntry);
    		keylen = curKeyEntry.getSize();
    		pos += keylen;
    		len -= keylen;
		}

		return new ServerNameExt(lengthExt, extdata);
	}
}
