package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.SignatureScheme;

public class SignatureAndHashAlgorithmExt extends TLSExtension
{
	public SignatureScheme[] data; // <2..2^16-2> bytes

	public SignatureAndHashAlgorithmExt(SignatureScheme[] sah) {
		type = 13; 
		this.data = sah;
	}

	public SignatureAndHashAlgorithmExt(int length, SignatureScheme[] sah) {
		this(sah); 
		this.setLength(length);
	}

	@Override
	public int getLengthFieldSize() {
		return 2;
	}

	@Override
	public int getDataSize() {
		if(data != null)
			return (data.length * 2);
		else
			return 0;
	}

	@Override
	public void toByteArrayData(byte[] msg, int position) {
		if(data != null) {
			for (int j=0; j < data.length; j++) {
				msg[position++] = (byte)(data[j].getValue() >>> 8);
				msg[position++] = (byte)(data[j].getValue());
			}
		}
	}

	@Override
	public void printData() {
		DebugInfo.printMessage(1,"  SignatureScheme:");
		if(data != null) {
                	for(int j=0; j<data.length; j++)
				DebugInfo.printMessage(1,"    " + data[j].toString());
		}
	}

	public static TLSExtension parseByteArray(int lengthExt, byte[] msg, int pos) {
		int len;
		int lengthFieldSize = 2;
    	if(lengthExt < lengthFieldSize) {
    		DebugInfo.printMessage(4,"===Error! SignatureAndHashAlgorithmExt::parseByteArray: Extension length failed");
    		return null;
    	}
		len = ConvertTypes.ntohs(msg,pos);
		pos += lengthFieldSize;

    	if(lengthExt != lengthFieldSize + len) {
    		DebugInfo.printMessage(4,"===Error! SignatureAndHashAlgorithmExt::parseByteArray: Extension length wrong");
    		return null;
    	}
    	
		if(len % 2 != 0) {
    		DebugInfo.printMessage(4,"===Error! SignatureAndHashAlgorithmExt::parseByteArray:  length must be even");
    		return null;
		}
		
		SignatureScheme[] extdata = new SignatureScheme[len/2];
    	for(int i=0; i<extdata.length; i++) {
    		extdata[i] = SignatureScheme.valueOf(ConvertTypes.ntohs(msg, pos));
    		pos += 2;
    	}

//		params = set cookie;
		
		return new SignatureAndHashAlgorithmExt(lengthExt, extdata);
	}
}
