package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.ProtocolVersion;

public class SupportedVersionsExt extends TLSExtension
{
	public int HelloMsgType; // <2..254> bytes
	public ProtocolVersion[] data; // <2..254> bytes

	public SupportedVersionsExt(ProtocolVersion version) {
		type = 43;
		HelloMsgType = 2;
		this.data = new ProtocolVersion[1];
		this.data[0] = version;
	}

	public SupportedVersionsExt(ProtocolVersion[] versions) {
		type = 43; 
		HelloMsgType = 1;
		this.data = versions;
	}

	public SupportedVersionsExt(int length, ProtocolVersion[] versions) {
		this(versions); 
		this.setLength(length);
	}

	public SupportedVersionsExt(int length, ProtocolVersion[] versions, int HelloMsgType) {
		this(length,versions);
		if(HelloMsgType <= 3)
			this.HelloMsgType = HelloMsgType;
	}

	@Override
	public int getLengthFieldSize() {
		if(this.HelloMsgType == 1)
			return 1;
		else
			return 0;
	}

	public boolean isTLS13() {
		if(data != null)  {
			for(int i = 0; i<data.length; i++) {
				if(data[i].isTLS13()) return true;
			}
		}
		return false;
	}

	@Override
	public int getDataSize() {
		if(data != null)
			return (data.length * 2);
		else
			return 0;
	}

	@Override
	public void toByteArrayData(byte[] msg, int position) {
		if(data != null) {
			for (int j=0; j < data.length; j++) {
				msg[position++] = (byte)(data[j].major);
				msg[position++] = (byte)(data[j].minor);
			}
		}
	}

	@Override
	public void printData() {
		String res = "  ";
		if(data != null) {
                	for(int j=0; j<data.length; j++)
				res += data[j].toString() + " ";
		}
		DebugInfo.printMessage(1,res);
	}

	public static TLSExtension parseByteArray(int lengthExt, byte[] msg, int pos, int HelloMsgType) {
		int len = 2;
		int lengthFieldSize = 0; // ServerHello
		if(HelloMsgType == 1) { lengthFieldSize = 1; len = msg[pos]; }// ClientHello
    	if(lengthExt < lengthFieldSize) {
    		DebugInfo.printMessage(4,"===Error! SupportedVersionsExt::parseByteArray: Extension length failed");
    		return null;
    	}
		pos += lengthFieldSize;

    	if(lengthExt != lengthFieldSize + len) {
    		DebugInfo.printMessage(4,"===Error! SupportedVersionsExt::parseByteArray: Extension length wrong");
    		return null;
    	}
    	
		if(len % 2 != 0) {
    		DebugInfo.printMessage(4,"===Error! SupportedGroupsExt::parseByteArray:  length must be even");
    		return null;
		}
		
		ProtocolVersion[] extdata = new ProtocolVersion[len/2];
    	for(int i=0; i<extdata.length; i++) {
    		extdata[i] = new ProtocolVersion(msg[pos++], msg[pos++]);
    	}

//		params = set cookie;
		
		return new SupportedVersionsExt(lengthExt, extdata, HelloMsgType);
	}
}
