package ru.ispras.telecomtesk.tlsv13.Mutation;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.Mutation.MutateField;

public class MutateOperator
{
	public int fieldCounter;
	public int valueCounter;
	public Field[] flds;
	public boolean mutationComplete;

	static Byte[] byteMutateValues = {0, 10, (byte)0xff};
	static Integer[] intMutateValues = {Integer.MIN_VALUE, 0, 10, 0xffffffff};
	
	public MutateOperator(Class<?> cls) {
		fieldCounter = 0;
		valueCounter = 0;
		mutationComplete = false;
		flds = cls.getFields();
		DebugInfo.printMessage(4,"MutateOperator:Class:getFields: " + flds.length + "fields found");
	}

	public MutateOperator(Object a) {
		fieldCounter = 0;
		valueCounter = 0;
		mutationComplete = false;
		flds = a.getClass().getFields();
		DebugInfo.printMessage(4,"MutateOperator:Object:getFields: " + flds.length + "fields found");
	}

	public static MutateField findAnnotation(Field fld) {
		Annotation[] annots = fld.getDeclaredAnnotations();
		if (annots == null) {
			return null;
		}
		for (Annotation ann : annots) {
			if (ann instanceof MutateField) {
				return (MutateField) ann;
			}
		}
		return null;
	}

	public Byte MutateByteField()
	{
		Byte curValue = null;
		
		System.out.println("It's byte type!");
		if(valueCounter < byteMutateValues.length) {
			curValue = byteMutateValues[valueCounter];
			System.out.println("current Byte mutate value: " + curValue.byteValue());
			valueCounter++;
		} else
			valueCounter = 0;
		
		return curValue;
	}
	
	public Integer MutateIntegerField()
	{
		Integer curValue = null;
		
		System.out.println("It's integer type!");
		if(valueCounter < intMutateValues.length) {
			curValue = intMutateValues[valueCounter];
			System.out.println("current Integer mutate value: " + curValue.intValue());
			valueCounter++;
		} else
			valueCounter = 0;

		return curValue;
	}
	
	public byte[] MutateByteArrayField()
	{
		byte[] curValue = null;
		System.out.println("It's byte[] type!");
		valueCounter = 0;
		return curValue;
	}
	
	public short[] MutateShortArrayField()
	{
		short[] curValue = null;
		System.out.println("It's short[] type!");
		valueCounter = 0;
		return curValue;
	}
	
	public Object getMutateFieldValue() {
		Object curFieldValue = null;

		System.out.println("getMutateFieldValue: valueCounter " + valueCounter);

		if (flds[fieldCounter].getType() == Byte.TYPE) {
			curFieldValue = MutateByteField();
		} else if (flds[fieldCounter].getType() == Integer.TYPE) {
			curFieldValue = MutateIntegerField();
		} else if (flds[fieldCounter].getType() == byte[].class) {
			curFieldValue = MutateByteArrayField();
		} else if (flds[fieldCounter].getType() == short[].class) {
			curFieldValue = MutateShortArrayField();
		} else {
			System.out.println("getMutateFieldValue: unknown Type " + flds[fieldCounter].getType());
		}

		return curFieldValue;
	}

	public int Mutate(Object a) {
		int res = -1;
		Object valueField = null;
		MutateField ann = null;
		
		DebugInfo.printMessage(4,"============");
		DebugInfo.printMessage(4,"Mutate start");

		if(flds == null) return 1;
		
		for(;fieldCounter < flds.length; fieldCounter++) {
			DebugInfo.printMessage(4,"------------");
			System.out.println("Field " + fieldCounter + ": " + flds[fieldCounter].getName() + ", Type: " + flds[fieldCounter].getType());
			ann = findAnnotation(flds[fieldCounter]);
			if (ann == null) {
				continue;
			}
			System.out.println("found comment: " + ann.comment());
			
			for(;;) {
				System.out.println("try find value");
				valueField = getMutateFieldValue();
				if (valueField == null) {
					break;
				}
				try {
					System.out.println("try set value");
					flds[fieldCounter].set(a, valueField);
					System.out.println("set value success");
					break;
				}
				catch (Exception err) {
					continue;
				}
			}
			
			if (valueField == null)
				continue;
			else 
				break;
		}
		
		System.out.println("fieldCounter " + fieldCounter);
		DebugInfo.printMessage(4,"============");
		if(fieldCounter >= flds.length) {
			mutationComplete = true;
			return 0;
		}

		return res;
	}
}