package ru.ispras.telecomtesk.tlsv13.TLSAlgorithms;

import java.math.BigInteger;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;

public enum NamedGroup {
        /* Elliptic Curve Groups (ECDHE) */
	SECP256R1(0X0017), 
	SECP384R1(0X0018),
	SECP521R1(0X0019),
	X25519(0X001D),
	X448(0X001E),

        /* Finite Field Groups (DHE) */
	FFDHEtest(0X01ff),
	FFDHE2048(0X0100),
	FFDHE3072(0X0101),
	FFDHE4096(0X0102),
	FFDHE6144(0X0103),
	FFDHE8192(0X0104),

        /* Reserved Code Points */
	/* ffdhe_private_use(0x01FC..0x01FF) */
        /* ecdhe_private_use(0xFE00..0xFEFF) */
	UNKNOWN(0xFFFF);
	
	public int group;
	
	private NamedGroup(int group) {
		this.group = group;
	}
	public int getValue(){
		return group;
	}

	public boolean isDHE(){
		switch (this) {
		case FFDHEtest: 
		case FFDHE2048: 
		case FFDHE3072: 
		case FFDHE4096: 
		case FFDHE6144: 
		case FFDHE8192: 
			return true;
		default: return false;
		}
	}
	public boolean isECDHE(){
		switch (this) {
		case SECP256R1:
		case SECP384R1:
		case SECP521R1:
			return true;
		default: return false;
		}
	}

	public static NamedGroup valueOf(int group) {
		switch (group) {
		case 0X0017: return SECP256R1;
		case 0X0018: return SECP384R1;
		case 0X0019: return SECP521R1;
		case 0X001D: return X25519;
		case 0X001E: return X448;
		case 0X0100: return FFDHE2048;
		case 0X0101: return FFDHE3072;
		case 0X0102: return FFDHE4096;
		case 0X0103: return FFDHE6144;
		case 0X0104: return FFDHE8192;
		case 0X01ff: return FFDHEtest;
		default: return UNKNOWN;
		}
	}

	public String toString() {
		return ConvertTypes.codeToString(this.group,this.name());
	}

	public static String toString(int group) {
		return ConvertTypes.codeToString(group,NamedGroup.valueOf(group).name());
	}

	public BigInteger getDHEgenerator () {
		switch (this) {
		case FFDHEtest: 
		case FFDHE2048: 
		case FFDHE3072: 
		case FFDHE4096: 
		case FFDHE6144: 
		case FFDHE8192: 
			return new BigInteger("2", 16);
		default: return null;
		}
	}

	public BigInteger getDHEmodulus () {
		switch (this) {
		case FFDHEtest:  
		case FFDHE2048: return TLS13SpecValues.p2048; 
		case FFDHE3072: return TLS13SpecValues.p3072;
		case FFDHE4096: return TLS13SpecValues.p4096;
		case FFDHE6144: return TLS13SpecValues.p6144;
		case FFDHE8192: return TLS13SpecValues.p8192;
		default: return null;
		}
	}
	
	public String getECname() {
		switch (this) {
		case SECP256R1:
			return "secp256r1";
		case SECP384R1:
			return "secp384r1";
		case SECP521R1:
			return "secp521r1";
		default: return "";
		}
	}

}
