package ru.ispras.telecomtesk.tlsv13.TLSSocket;

import java.io.*;
import java.util.ArrayList;
import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

public class TLSResponse
{
    byte[][] message = new byte[10][];
    int lastind;
    int lastpos;
    int curind;
    int curpos;
    // responseTime -   (       )
    public long responseTime = -1;
    int TLSRESPONSE_ARRAY_LENGTH = 65565;
    public int err;
    String errstr;
    
    public TLSResponse() {
    	lastind = 0; lastpos = 0; curind = 0;
    	curpos = 0; errstr = null; responseTime = -1;
    	message[0] = new byte[TLSRESPONSE_ARRAY_LENGTH];
    }
    
    public TLSResponse(byte[] tlsRaw) {
    	lastind = 0; lastpos = 0; curind = 0;
    	curpos = 0; errstr = null; responseTime = -1;
    	if(tlsRaw != null) {
    		message[0] = tlsRaw;
    		lastpos = tlsRaw.length;
    	} else
    		message[0] = new byte[TLSRESPONSE_ARRAY_LENGTH];
    }
    
    public void init() {
    	lastind = 0; lastpos = 0; curind = 0;
    	curpos = 0; errstr = null; responseTime = -1;
    }

    public void shiftPosition(int n) {
		DebugInfo.printMessage(5,"TLSResponse::shiftPosition");
    	lastpos += n;
    	
		if(lastpos > message[lastind].length)
			throw new ArrayIndexOutOfBoundsException( "TLSResponse::shiftCurpos: bad curpos value");
		else if(lastpos == message[lastind].length) {
	    	lastpos = 0;
			lastind++;
			if(message[lastind] == null)
				message[lastind] = new byte[TLSRESPONSE_ARRAY_LENGTH];
		}
    }

    public int readbytes(DataInputStream instream) throws IOException {
		return instream.read(message[lastind],lastpos,message[lastind].length - lastpos);
    }

    public void printResponse() {
    	if(lastind == 0 && lastpos == 0)
    		DebugInfo.printMessage(6, "Socket response: no data");
    	else
        	DebugInfo.printMessage(6, "Socket response:");
    	
    	for(int i=0;i<=lastind;i++) {
    		if(i==lastind)
    			DebugInfo.printDump(6, message[i], 0, lastpos, null);
    		else
    			DebugInfo.printDump(6, message[i], null);
    	}
    }

    public int getFullLength() {
    	int len=0;
    	for(int i=0; i<lastind; i++)
    		len += message[i].length;
    	len += lastpos;
    	return len;
    }

    private int getRemainLength() {
    	int len=0;
    	if((lastind >= 0)&&(curind >= 0)&&(curind <= lastind)) {
    		while(curind < lastind) {
    			len += message[curind].length;
    			curind++;
    		}
    		if(lastpos >= 0) len += lastpos;
    		if(curpos >= 0) len -= curpos;
    	}
    	if(len < 0) len = 0;
    	return len;
    }

    // get 'length' bytes from TLSResponse.message[curind][curpos]
    private byte[] getBytesArray(int length) {
    	int remainlen = getRemainLength();
    	int len, copylen, dstpos = 0;
    	byte[] a = null;
    	
		if(length >= 0) {
    		copylen = (length <= remainlen) ? length : remainlen;
    		a = new byte[copylen];
    		
    		while(copylen > 0) {
				len = message[curind].length - curpos;
				if(len <= a.length - dstpos) {
					System.arraycopy(message[curind], curpos, a, dstpos, len);
					curind++; 
					if(curpos != 0) curpos=0;
					dstpos += len;
					copylen -= len;
				}
				else {
					len = a.length - dstpos;
					System.arraycopy(message[curind], curpos, a, dstpos, len);
					curpos += len;
					dstpos += len;
					copylen -= len;
				}
    		}
    	}
		return a;
    }

    // write 'length' bytes from TLSResponse.message[curind][curpos] to array a
    private int getBytesArray(int length, byte[] a) {
    	int remainlen = getRemainLength();
    	int len, res, copylen, dstpos = 0;
    	
    	res = 0;
		if(remainlen == 0) res = -1;
		else if(length >= 0  &&  a != null  &&  length <= a.length) {
    		copylen = (length <= remainlen) ? length : remainlen;
    		res = copylen;
    		
    		while(copylen > 0) {
				len = message[curind].length - curpos;
				if(len <= a.length - dstpos) {
					System.arraycopy(message[curind], curpos, a, dstpos, len);
					curind++; 
					if(curpos != 0) curpos=0;
					dstpos += len;
					copylen -= len;
				}
				else {
					len = a.length - dstpos;
					System.arraycopy(message[curind], curpos, a, dstpos, len);
					curpos += len;
					dstpos += len;
					copylen -= len;
				}
    		}
    	}
		return res;
    }

    // cut TLSResponse.message into TLS messages and create ArrayList
    public ArrayList<RecordLayer> getTlsArrayCrypto() {
    	ArrayList<RecordLayer> tlslst = new ArrayList<RecordLayer>();
    	byte[] tlshdr = new byte[5];
    	byte[] tlsdata;
    	RecordLayer newtls;
    	TLSContentType tlstype;
    	int res,len;
    	err = 0; curind = 0; curpos = 0;
    	    	
    	if(getFullLength() > 0) {
    		while((res = getBytesArray(5,tlshdr)) >= 0) {
    			if(res != 5) {
    				err = -1;
    				errstr = "===Error! TLSResponse::getTlsArrayCrypto: Failed read TLS header";
    				break;
    			}
    			tlstype = TLSContentType.valueOf(tlshdr[0]);
    			newtls = new RecordLayer(tlstype, new ProtocolVersion(tlshdr[1], tlshdr[2]));
    	    	len = ConvertTypes.ntohs(tlshdr,3);
    			newtls.setLength(len);
    			System.out.println("Array length " + len);
    			tlsdata = getBytesArray(len);
    			if(tlsdata == null  ||  tlsdata.length != len) {
    				err = -2;
    				errstr = "===Error! TLSResponse::getTlsArrayCrypto: Failed read TLS data";
    				break;
    			}
    			newtls.setNewContent(tlsdata);
    			tlslst.add(newtls);
    		}
    	}

    	if(err < 0) {
    		System.out.println(errstr);
    	}

    	return tlslst;
    }

    // decrypt TLS messages ArrayList
    public void getTlsArrayPlain(ArrayList<RecordLayer> cryptomsg, TLSConnection con) throws Exception {
    	RecordLayer newtls;
    	
    	if((cryptomsg != null)&&(con != null)) {
    		for(int i=0; i < cryptomsg.size(); i++) {
    			newtls = cryptomsg.get(i);
    			err = newtls.getTlsPlain(con);
    			if(err != 0) break;
    		}
    	}
    }
    
    // check and trim Padding
    public void trimPadding(ArrayList<RecordLayer> cryptomsg) throws Exception {
    	RecordLayer newtls;
    	
    	if(cryptomsg != null) {
    		for(int i=0; i < cryptomsg.size(); i++) {
    			newtls = cryptomsg.get(i);
    			err = newtls.trimPadding();
    			if(err != 0) break;
    		}
    	}
    }
    
    // parse TLS messages ArrayList
    public void parseTlsArrayPlain(ArrayList<RecordLayer> plainmsg, TLSConnection con) {
    	RecordLayer newtls;
    	err = 0;
    	int i;
    	
    	if(plainmsg != null) {
    		for(i=0; i < plainmsg.size() && (err == 0); i++) {
    			newtls = plainmsg.get(i);
    			err = newtls.parseTlsPlain(con);
    		}
    	}
    }

    public void parseResponse(TLSConnection con) throws Exception {
    	err=0;
    	ArrayList<RecordLayer> response_msgs;
    	
        // cut TLSResponse.message into TLS messages and create ArrayList
    	response_msgs = getTlsArrayCrypto();
        // decrypt TLS messages ArrayList
    	if(err == 0)
    		getTlsArrayPlain(response_msgs, con);
        // parse TLS messages ArrayList
    	if(err == 0)
    		parseTlsArrayPlain(response_msgs, con);
    }

}
