package ru.ispras.telecomtesk.tlsv13.TLSserverSpecification;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.security.cert.CertificateFactory;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.io.*;

import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSSocket.*;
import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.TLSspecification.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TlsKeystore;

public class TLSServerSpec
{
	public TLSSocket sock = null;
	public TLSScenarioState currentState = TLSScenarioState.DISCONNECTED;
	public TLSScenarioState previousState = TLSScenarioState.DISCONNECTED;
	protected int checkHeader = -1;
	protected boolean isOutOfOrder = false;
	protected String strerr;
	protected Hello last_client_hello = null;
	protected boolean isFixedDH = false;
	public boolean isCertificateVerifyMustSend = false;
	public boolean isServerMustSendCertificate = true;
	public boolean afterServerFinished  = false;
	public boolean isHandshakeCompleted  = false;
	protected boolean isChangeCipher = false;
	public boolean mayNoResponse = false;
	public boolean mustNoResponse = false;
	public boolean mustSendEndOfEarlyData = false;
	protected ClientCertificateType clientCertType = ClientCertificateType.UNKNOWN;
	public boolean isSessionResume = false;
	public boolean isGetCertificateRequest = false;
	public boolean isClientPostAuth = false;
	public boolean noRenegotiation = false;
	public boolean secure_renegotiation = false;
	public boolean initial_handshake = true;
	public boolean needCookie = false;
	public MessageDirection curMsgDir;
	public int waitFatalAlert = -1;
	public int disconnectWithSendAlert = -1;
	public long responseTime = -1;
	public ArrayList sessionMsgList = new ArrayList();
	public ArrayList<NewSessionTicket> sessionTicketsList = new ArrayList<NewSessionTicket>();
	public ArrayList<TicketPSK> sessionTicketsPSK = new ArrayList<TicketPSK>();
	public int last_client_hello_index = 0;
	public X509Certificate serverCert = null;
	public Certificate localCert = null;
	public byte[] pre_master_secret = null;
	public byte[] client_verify_data = null;
	public byte[] server_verify_data = null;

	public TLSServerSpec() {
	}
	
	public TLSScenarioState getCurrentState() {
		return currentState;
	}
	
	public void setCurrentState(TLSScenarioState value) {
		this.currentState = value;
	}

	public boolean isHandshakeComplete() {
		return isHandshakeCompleted;
	}

	public void setClientCertificateType() {
		this.clientCertType = ClientCertificateType.valueOf(TlsKeystore.getSigAlg(localCert));
	}

	public void printFlags(int debugLevel) {
		DebugInfo.printMessage(debugLevel, "TLSServerSpec:");
		DebugInfo.printMessage(debugLevel, "   currentState: " + currentState.name());
		DebugInfo.printMessage(debugLevel, "   previousState: " + previousState.name());
		DebugInfo.printMessage(debugLevel, "   checkHeader: " + checkHeader);
		DebugInfo.printMessage(debugLevel, "   isOutOfOrder: " + isOutOfOrder);
		DebugInfo.printMessage(debugLevel, "   strerr: " + strerr);
		DebugInfo.printMessage(debugLevel, "   isFixedDH: " + isFixedDH);
		DebugInfo.printMessage(debugLevel, "   isCertificateVerifyMustSend: " + isCertificateVerifyMustSend);
		DebugInfo.printMessage(debugLevel, "   isServerMustSendCertificate: " + isServerMustSendCertificate);
		DebugInfo.printMessage(debugLevel, "   isGetCertificateRequest: " + isGetCertificateRequest);
		DebugInfo.printMessage(debugLevel, "   needCookie: " + needCookie);
		DebugInfo.printMessage(debugLevel, "   noRenegotiation: " + noRenegotiation);
		DebugInfo.printMessage(debugLevel, "   secure_renegotiation: " + secure_renegotiation);
		DebugInfo.printMessage(debugLevel, "   initial_handshake: " + initial_handshake);
		DebugInfo.printMessage(debugLevel, "   isSessionResume: " + isSessionResume);
		DebugInfo.printMessage(debugLevel, "   isHandshakeCompleted: " + isHandshakeCompleted);
		DebugInfo.printMessage(debugLevel, "   afterServerFinished: " + afterServerFinished);
		DebugInfo.printMessage(debugLevel, "   isClientPostAuth: " + isClientPostAuth);
		DebugInfo.printMessage(debugLevel, "   isChangeCipher: " + isChangeCipher);
		DebugInfo.printMessage(debugLevel, "   mayNoResponse: " + mayNoResponse);
		DebugInfo.printMessage(debugLevel, "   mustNoResponse: " + mayNoResponse);
		DebugInfo.printMessage(debugLevel, "   mustSendEndOfEarlyData: " + mayNoResponse);
		DebugInfo.printMessage(debugLevel, "   clientCertType: " + clientCertType);
		DebugInfo.printMessage(debugLevel, "   curMsgDir: " + curMsgDir);
		DebugInfo.printMessage(debugLevel, "   waitFatalAlert: " + waitFatalAlert);
		DebugInfo.printMessage(debugLevel, "   disconnectWithSendAlert: " + disconnectWithSendAlert);
		DebugInfo.printMessage(debugLevel, "   responseTime: " + responseTime);
		DebugInfo.printMessage(debugLevel, "   last_client_hello_index: " + last_client_hello_index);
		DebugInfo.printMessage(debugLevel, "   sessionMsgList size: " + sessionMsgList.size() + "\n");
	}
	
	public void init(TLSConnection con) throws IOException {
		DebugInfo.printMessage(1,"spec flags set to default values");

		checkHeader = -1;
		isOutOfOrder = false;
		last_client_hello = null;
		isFixedDH = false;
		isCertificateVerifyMustSend = false;
		isServerMustSendCertificate = true;
		isGetCertificateRequest = false;
		noRenegotiation = false;
		secure_renegotiation = false;
		initial_handshake = true;
		isSessionResume = false;
		isHandshakeCompleted = false;
		afterServerFinished = false;
		isClientPostAuth = false;
		isChangeCipher = false;
		needCookie = false;
		mayNoResponse = false;
		mustNoResponse = false;
		mustSendEndOfEarlyData = false;
		clientCertType = ClientCertificateType.UNKNOWN;
		waitFatalAlert = -1;
		disconnectWithSendAlert = -1;
		responseTime = -1;
		previousState = currentState;
		sessionMsgList.clear();
//		sessionTicketsList.clear();
//		sessionTicketsPSK.clear();
		last_client_hello_index = 0;
		serverCert = null;
		localCert = null;
		pre_master_secret = null;
		client_verify_data = null;
		server_verify_data = null;
		
		con.init();
		return;
	}

	public void connect(TLSConnection con) throws IOException {
		DebugInfo.printMessage(1,"connect state");

		if(sock != null) {
			sock.connect();
		}
		
		this.init(con);

		setCurrentState(TLSScenarioState.CONNECTED);
			
		System.out.println("State = " + currentState + "\n");
		return;
	}

	public void disconnect() {
		DebugInfo.printMessage(3,"TLSServerSpec::disconnect");

		if(sock != null) {
			sock.disconnect();
		}
		isHandshakeCompleted = false;
			
		previousState = currentState;
		setCurrentState(TLSScenarioState.DISCONNECTED);
		System.out.println("State = " + currentState + "\n\n");
	}
	
	
	public ArrayList<RecordLayer> wait(long time, TLSConnection con) throws Exception {
		DebugInfo.printMessage(3,"TLSServerSpec::wait");

		if(!currentState.connected) 
			return null;
		
		DebugInfo.printMessage(3,"TLSServerSpec::wait start");

		disconnectWithSendAlert = -1;
		isOutOfOrder = false;
		mayNoResponse = true;
		
		TLSResponse resp = waitTLS(time, con);

    	if(resp.err != 0)
    		return null;
    
    	return receiveTLS(resp, con);	
    }
	
	public TLSResponse waitTLS(long time, TLSConnection con) throws Exception {
			System.out.println("waitTLS mediator");
			long prevTimeout = sock.getTimeout();
			sock.setTimeout(time);
			sock.waitForResponse();
			TLSResponse result = sock.getResponse(); 
			sock.setTimeout(prevTimeout);
			System.out.println("waitTLS mediator return");
			return result;
	}
		
	public ArrayList<RecordLayer> send(RecordLayer tlsmsg, TLSConnection con) throws Exception {
		DebugInfo.printMessage(3,"TLSServerSpec::send");

		if((!currentState.connected) || 
			(tlsmsg == null) || (tlsmsg.content == null) || (tlsmsg.content[0] == null))
				return null;

		curMsgDir = MessageDirection.getMessageDirection(con.localRole, TLSDirection.SEND);
		
		DebugInfo.printMessage(3,"TLSServerSpec::send 2");
		printFlags(5);
		con.getSecurityParams().printDump(6);

		disconnectWithSendAlert = -1;
		
		sendPreSet(tlsmsg, con);

		printFlags(5);
		con.getSecurityParams().printDump(6);
		
		isOutOfOrder = false;

		TLSResponse resp = sendTLS(tlsmsg, con);
		return receiveTLS(resp, con);	
    }
		
		
	public ArrayList<RecordLayer> send(ArrayList<RecordLayer> tlsmsglst, TLSConnection con) throws Exception {
		DebugInfo.printMessage(3,"TLSServerSpec::send ArrayList");

		if((!currentState.connected) || 
			(tlsmsglst == null) || (tlsmsglst.size() == 0))
				return null;

		for(RecordLayer rlmsg : tlsmsglst) {
			if((rlmsg == null) || (rlmsg.content == null) || (rlmsg.content[0] == null))
				return null;
		}
		
		curMsgDir = MessageDirection.getMessageDirection(con.localRole, TLSDirection.SEND);

		DebugInfo.printMessage(3,"TLSServerSpec::send ArrayList 2");
		printFlags(5);
		con.getSecurityParams().printDump(6);
		
		disconnectWithSendAlert = -1;

		for(RecordLayer tlsmsg : tlsmsglst) {
			sendPreSet(tlsmsg, con);
			if(waitFatalAlert >= 0  ||  currentState == TLSScenarioState.DISCONNECTED) 
				break;
		}
		
		System.out.println("send pre comlete\n");
		printFlags(5);
		con.getSecurityParams().printDump(6);
		
		isOutOfOrder = false;

		TLSResponse resp = sendTLS(tlsmsglst, con);
 		return receiveTLS(resp, con);	
    }
		
		
	private void sendPreSet(RecordLayer tlsmsg, TLSConnection con) throws Exception {
		System.out.println("sendPreSet start");
		
		DebugInfo.printMessage(6,"TLSServerSpec::send: increment Sequence Number");
		con.incrementSenderSequenceNumber();
		
		if(tlsmsg.msglist.size() == 0)
			System.out.println("===Warning! TLSServerSpec::send: Message list is empty");

		for(Object msg : tlsmsg.msglist) {
				
			if(msg instanceof HandshakeMessage) {
				System.out.println("HandshakeMessage start");
				if(msg instanceof HelloRetryRequest)
					sendHelloRetryRequest((HelloRetryRequest)msg);
				else if(msg instanceof Hello) {
					System.out.println("Hello start");
					if(((HandshakeMessage) msg).msg_type == 1)
						sendClientHello((Hello)msg,con);
					else
						sendServerHello((Hello)msg);
				}
				else if(msg instanceof HelloRequest)
					sendHelloRequest((HelloRequest)msg);
				else if(msg instanceof HMCertificate)
					sendCertificate((HMCertificate)msg,con);
				else if(msg instanceof CertificateRequest)
					sendCertificateRequest((CertificateRequest)msg);
				else if(msg instanceof ClientKeyExchange)
					sendClientKeyExchange((ClientKeyExchange)msg);
				else if(msg instanceof ServerKeyExchange)
					sendServerKeyExchange((ServerKeyExchange)msg);
				else if(msg instanceof ServerHelloDone)
					sendServerHelloDone((ServerHelloDone)msg);
				else if(msg instanceof EndOfEarlyData)
					sendEndOfEarlyData((EndOfEarlyData)msg);
				else if(msg instanceof CertificateVerify)
					sendCertificateVerify((CertificateVerify)msg,con);
				else if(msg instanceof EncryptedExtensions)
					sendEncryptedExtensions((EncryptedExtensions)msg);
				else if(msg instanceof Finished)
					sendFinished((Finished)msg, con);
				else if(msg instanceof NewSessionTicket)
					sendNewSessionTicket((NewSessionTicket)msg);
				else if(msg instanceof KeyUpdate)
					sendKeyUpdate((KeyUpdate)msg, con);
				else {
					if(tlsmsg.tls_type == TLSContentType.HANDSHAKE)
						sendHandshakeUndefinedType((HandshakeUndefinedType)msg);
					else {
						System.out.println("===Warning! TLSServerSpec::send: Unknown message type: " + tlsmsg.tls_type);
						sendUndefinedType();
					}
				}
				sessionMsgList.add(msg);
				con.hskParams.sessionMsgList.add(msg);
			}
			else if(msg instanceof Alert) {
				sendAlert((Alert)msg);
			}
			else if(msg instanceof ChangeCipher) {
				sendChangeCipher((ChangeCipher)msg);
				// ChangeCipher must not encrypted
				con.decrementSenderSequenceNumber();
			}
			else if(msg instanceof ApplicationData) {
				sendApplicationData((ApplicationData)msg);
			}
			else {
				System.out.println("===Warning! TLSServerSpec::send: Unknown message type: " + tlsmsg.tls_type);
				sendUndefinedType();
			}
				
			if(waitFatalAlert >= 0  ||  currentState == TLSScenarioState.DISCONNECTED) 
				break;
		}
		
		System.out.println("sendPreSet comlete\n");
    }
		
		
	public TLSResponse sendTLS(RecordLayer tlsmsg, TLSConnection con) throws Exception {
		TLSResponse result = null;
		DebugInfo.printMessage(3,"sendTLS mediator");
		if(sock != null) {
			sock.sendMessage(tlsmsg.toByteArray());
			result = sock.getResponse(); 
		} else {
			DebugInfo.printMessage(1,"===Error! no socket for send\n");
		}
		DebugInfo.printMessage(3,"sendTLS mediator return");
		return result;
	}

	public TLSResponse sendTLS(ArrayList<RecordLayer> tlsmsglst, TLSConnection con) throws Exception {
		int i,len = 0;
		RecordLayer tlsmsg;
		System.out.println("sendTLS ArrayList mediator");

		for(i=0; i<tlsmsglst.size(); i++) {
			len += tlsmsglst.get(i).getSize();
		}
		System.out.println("sendTLS len " + len);
		byte[] res = new byte[len];
		for(i=0,len=0; i<tlsmsglst.size(); i++) {
			tlsmsg = tlsmsglst.get(i);
			System.arraycopy(tlsmsg.toByteArray(),0,res,len,tlsmsg.getSize());
			len += tlsmsg.getSize();
		}
		System.out.println("sendTLS array created");
		
		sock.sendMessage(res);
		System.out.println("sendTLS send");
		TLSResponse result = sock.getResponse(); 
		System.out.println("sendTLS ArrayList mediator return");
		return result;
	}

	public ArrayList<RecordLayer> receiveTLS(TLSResponse resp, TLSConnection con) throws Exception {	
		System.out.println("\nTLSServerSpec::receiveTLS");
		if((resp == null) || (resp.err != 0))
			return null;
		
		ArrayList<RecordLayer> response_msgs = null;

		curMsgDir = MessageDirection.getMessageDirection(con.localRole, TLSDirection.RECEIVE);

		// cut response to TLS records
		response_msgs = cutResponse(resp, con);
		mayNoResponse = false;
		DebugInfo.printMessage(5,"TLSServerSpec::receiveTLS cutResponse ok");
    	if(disconnectWithSendAlert != -1) {
			DebugInfo.printMessage(5,"TLSServerSpec::receiveTLS disconnectWithSendAlert " + disconnectWithSendAlert);
//    		return null;
    	}
    	else if(response_msgs.size() == 0)
    		return response_msgs;
    	else if(currentState == TLSScenarioState.DISCONNECTED) {
			DebugInfo.printMessage(1,"===Error! Receive response in state DISCONNECTED");
    		idlingDecrypt(response_msgs, 0, con);
    		return null;
    	}

		previousState = currentState;

		DebugInfo.printMessage(4,"TLSServerSpec::receive: parse start");
		for(RecordLayer msgrl : response_msgs) {

			disconnectWithSendAlert = msgrl.decryptAndParseTLS(con);
			
			DebugInfo.printMessage(6,"TLSServerSpec::receive: increment Sequence Number after decryption");
			con.incrementReceiverSequenceNumber();
		    	
            if( disconnectWithSendAlert != -1 ) {
        		DebugInfo.printMessage(1,"decryptAndParseTLS return " + disconnectWithSendAlert);
        		DebugInfo.printMessage(4,"waitFatalAlert " + waitFatalAlert);
                break ;
            }

            if( msgrl.version.minor != 3 ) {
            	if( (msgrl.tls_type == TLSContentType.HANDSHAKE) && 
            		(msgrl.msglist != null) && 
            		(msgrl.msglist.get(0) instanceof Hello) ) ; // ignore
            	else {
            		DebugInfo.printMessage(1,"===Error! Bad minor TLS version");
            		DebugInfo.printMessage(4,"waitFatalAlert " + waitFatalAlert);
            		break ;
            	}
            }

	    	KeyExchangeAlgorithm ke_alg = con.getSecurityParams().ke_alg;

			DebugInfo.printMessage(4,"TLSServerSpec::receive: check plain message");
			for(Object msg : msgrl.msglist) {

				if(msg instanceof HandshakeMessage) {
					if(msg instanceof HelloRetryRequest)
						receiveHelloRetryRequest((HelloRetryRequest)msg,con);
					else if(msg instanceof Hello)
						receiveHello((Hello)msg,con);
					else if(msg instanceof HelloRequest)
						receiveHelloRequest((HelloRequest)msg);
					else if(msg instanceof EncryptedExtensions)
						receiveEncryptedExtensions((EncryptedExtensions)msg,con);
					else if(msg instanceof HMCertificate) {
						try {
							receiveCertificate((HMCertificate)msg,con);
						}
						catch (Exception err) {}
					}
					else if(msg instanceof CertificateRequest)
						receiveCertificateRequest((CertificateRequest)msg,con);
					else if(msg instanceof ClientKeyExchange)
						receiveClientKeyExchange((ClientKeyExchange)msg);
					else if(msg instanceof ServerKeyExchange)
						receiveServerKeyExchange((ServerKeyExchange)msg,con.getSecurityParams());
					else if(msg instanceof ServerHelloDone)
						receiveServerHelloDone((ServerHelloDone)msg);
					else if(msg instanceof EndOfEarlyData)
						receiveEndOfEarlyData((EndOfEarlyData)msg);
					else if(msg instanceof CertificateVerify)
						receiveCertificateVerify((CertificateVerify)msg,con);
					else if(msg instanceof Finished)
						receiveFinished((Finished)msg, con);
					else if(msg instanceof NewSessionTicket)
						receiveNewSessionTicket((NewSessionTicket)msg, con);
					else if(msg instanceof KeyUpdate)
						receiveKeyUpdate((KeyUpdate)msg, con);
					else {
						System.out.println("===Error! TLSServerSpec::receive: Unknown Handshake message type: " + ((HandshakeMessage)msg).msg_type);
						disconnectWithSendAlert = 10; // unexpected_message
						break;
					}
					sessionMsgList.add(msg);
					con.hskParams.sessionMsgList.add(msg);
				}
				else if(msg instanceof Alert) {
					receiveAlert((Alert)msg);
				}
				else if(msg instanceof ChangeCipher) {
					receiveChangeCipher((ChangeCipher)msg);
					// ChangeCipher must not encrypted
					con.decrementReceiverSequenceNumber();
				}
				else if(msg instanceof ApplicationData) {
					receiveApplicationData((ApplicationData)msg);
				}
				else {
					System.out.println("===Error! TLSServerSpec::receive: Unknown message type " + msgrl.tls_type);
					disconnectWithSendAlert = 10; // unexpected_message
					break;
				}
					
				if(disconnectWithSendAlert >= 0  ||  currentState == TLSScenarioState.DISCONNECTED) 
					break;
			}
			if(disconnectWithSendAlert >= 0  ||  currentState == TLSScenarioState.DISCONNECTED) 
				break;
		}

		printFlags(5);
		DebugInfo.printMessage(5,"SecurityParams States\n");
		con.printDump(5);
		
		isOutOfOrder = false;
			
		return response_msgs;
	}

	ArrayList<RecordLayer> cutResponse(TLSResponse resp, TLSConnection con) throws Exception {
		ArrayList<RecordLayer> resTlsList = null;

		if(resp != null) {
			resTlsList = resp.getTlsArrayCrypto();
			responseTime = resp.responseTime;
			if(resTlsList.size() == 0) {
				DebugInfo.printMessage(3,"TLSServerSpec::cutResponse: Message list is empty");
				if(waitFatalAlert > 0) {
					waitFatalAlert = -1;
					disconnectWithSendAlert = 80;
					DebugInfo.printMessage(3,"Wait fatal alert, but no response");
				}
				else if(   currentState == TLSScenarioState.DISCONNECTED
				   || currentState == TLSScenarioState.AFTER_ClientCertificate
				   || currentState == TLSScenarioState.AFTER_ClientCertificateVerify
				   || currentState == TLSScenarioState.AFTER_ServerCloseNotify
				   || currentState == TLSScenarioState.AFTER_ClientFinished
				   || currentState == TLSScenarioState.AFTER_ServerFinished
				   || mayNoResponse
				   || mustNoResponse)
						;
				else if(currentState == TLSScenarioState.AFTER_ClientCloseNotify) {
					disconnectWithSendAlert = -1;
					DebugInfo.printMessage(3,"Wait CloseNotify alert, but no response");
				}
				else {
					disconnectWithSendAlert = 80;
					DebugInfo.printMessage(3,"Unexpected no response. State " + currentState);
				}
			} else {
				if(mustNoResponse) {
					DebugInfo.printMessage(2,"===Error! TLSServerSpec::cutResponse: must no response, but receive message");
					waitFatalAlert = -1;
					disconnectWithSendAlert = 10;
				}
			}

	    	if((waitFatalAlert > 0) && (resTlsList.get(0).tls_type != TLSContentType.ALERT)) {
				DebugInfo.printMessage(2,"===Error! TLSServerSpec::cutResponse: Alert message expected");
				waitFatalAlert = -1;
				disconnectWithSendAlert = 10;
				DebugInfo.printMessage(3,"Wait fatal alert, but receive other message");
	    	}
		}
       	return resTlsList;
	}
	
	void sendServerHello (Hello msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;

			previousState = currentState;
			setCurrentState(TLSScenarioState.AFTER_ServerHello);
	}


	void sendHelloRetryRequest (HelloRetryRequest msg) {
		isOutOfOrder = true;
		waitFatalAlert = 10;

		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ServerHelloRetry);
	}


	void sendClientHello (Hello msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
		}
		else {
			checkHeader = checkHello(msg);
			if(checkHeader >= 0) {
				System.out.println("TLSServerSpec::checkHello: " + strerr);
				waitFatalAlert = checkHeader;
			}
		}
		
		previousState = currentState;
		noRenegotiation = false;
		isSessionResume = false;
		last_client_hello = msg;
		last_client_hello_index = sessionMsgList.size() > 0 ? sessionMsgList.size() : 0;
			
		setCurrentState(TLSScenarioState.AFTER_ClientHELLO);

		con.setRandom(msg.rnd, true);
		con.updateState(msg);
}


	void sendHelloRequest (HelloRequest msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
			mayNoResponse = true;
	}

	void sendServerHelloDone (ServerHelloDone msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
			previousState = currentState;
			setCurrentState(TLSScenarioState.AFTER_ServerHelloDone);
	}
	
	void sendEncryptedExtensions (EncryptedExtensions msg) {
		isOutOfOrder = true;
		waitFatalAlert = 10;
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_EncryptedExtensions);
	}
	
	void sendEndOfEarlyData (EndOfEarlyData msg) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		}
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_EndOfEarlyData);
	}
	
	void sendServerKeyExchange (ServerKeyExchange msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
			previousState = currentState;
			setCurrentState(TLSScenarioState.AFTER_ServerKeyExchange);
	}

	void sendCertificateRequest (CertificateRequest msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
			previousState = currentState;
			setCurrentState(TLSScenarioState.AFTER_CertificateRequest);
	}

	void sendCertificate (HMCertificate msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		} else
			waitFatalAlert = CheckMessageHeader.checkHeader(msg);
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ClientCertificate);
	}

	 void sendClientKeyExchange (ClientKeyExchange msg) {
			isOutOfOrder = true;
			waitFatalAlert = 10;
			previousState = currentState;
			setCurrentState(TLSScenarioState.AFTER_ClientKeyExchange);
	}

	void sendCertificateVerify (CertificateVerify msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		} else
			waitFatalAlert = CheckMessageHeader.checkHeader(msg, con);
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ClientCertificateVerify);
	}

	void sendFinished (Finished msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		} else
			waitFatalAlert = CheckMessageHeader.checkHeader(msg, con, curMsgDir);
		
		previousState = currentState;
		if(con.localRole == ConnectionRole.CLIENT)
			setCurrentState(TLSScenarioState.AFTER_ClientFinished);
		else
			setCurrentState(TLSScenarioState.AFTER_ServerFinished);
		
		con.updateState(msg, curMsgDir);
		
		if((waitFatalAlert < 0) && (currentState == TLSScenarioState.AFTER_ClientFinished))
			isHandshakeCompleted = true;
	}

	void sendNewSessionTicket (NewSessionTicket msg) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		} else
			waitFatalAlert = CheckMessageHeader.checkHeader(msg);
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_NewSessionTicket);
	}

	void sendKeyUpdate (KeyUpdate msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
		} else
			waitFatalAlert = CheckMessageHeader.checkHeader(msg);
		
		previousState = currentState;
		if(con.localRole == ConnectionRole.CLIENT)
			setCurrentState(TLSScenarioState.AFTER_ClientKeyUpdate);
		else
			setCurrentState(TLSScenarioState.AFTER_ServerKeyUpdate);
		
		if(waitFatalAlert < 0)
			con.updateState(msg, curMsgDir);
	}

	void sendChangeCipher (ChangeCipher msg) {
		if(!CheckMessageOrder.checkOrder(msg, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);
	}

	void sendAlert (Alert msg) {
			System.out.println("TLSServerSpec::checkAlert: \n" + msg.toString());
			
			checkHeader = checkAlert(msg);
			if(checkHeader >= 0) {
				System.out.println("TLSServerSpec::checkAlert: " + strerr);
				waitFatalAlert = checkHeader;
			}

			previousState = currentState;

			if(msg.isCloseNotifyAlert()) {
				if(currentState == TLSScenarioState.AFTER_ClientCloseNotify) {
					isOutOfOrder = true;
					mustNoResponse = true;
				}
				else if(currentState == TLSScenarioState.AFTER_ServerCloseNotify) {
					setCurrentState(TLSScenarioState.DISCONNECTED);
					mustNoResponse = true;
				}
				else
					setCurrentState(TLSScenarioState.AFTER_ClientCloseNotify);
			}
			else 
				setCurrentState(TLSScenarioState.DISCONNECTED);
	}

	void sendApplicationData (ApplicationData msg) {

			if(  !isHandshakeCompleted
			  || currentState == TLSScenarioState.AFTER_ClientCloseNotify) {
				isOutOfOrder = true;
				waitFatalAlert = 10;
			}
			
			mayNoResponse = true;
	}

	void sendHandshakeUndefinedType (HandshakeUndefinedType msg) {
			System.out.println("===Warning! TLSServerSpec::send: Unknown Handshake message type: " + ((HandshakeMessage)msg).msg_type);
			System.out.println(msg.toString());
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
	}

	void sendUndefinedType () {
			isOutOfOrder = true;
			waitFatalAlert = 10; // unexpected_message
	}

	void receiveHelloRetryRequest(HelloRetryRequest msg, TLSConnection con) {
		if(currentState != TLSScenarioState.AFTER_ClientHELLO) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		else {
			checkHeader = checkHello(msg);
			if(checkHeader >= 0) {
				System.out.println("TLSServerSpec::receiveHelloRetryRequest:checkHello: " + strerr);
				disconnectWithSendAlert = checkHeader;
			}
		}

		needCookie = true;
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ServerHelloRetry);
		
		if(initial_handshake)
			initial_handshake = false;
					
		con.updateState(msg);
	}
	
	void receiveHello(Hello msg, TLSConnection con) {
		if(msg.msg_type != 2  
		  || currentState != TLSScenarioState.AFTER_ClientHELLO) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		else {
			checkHeader = checkHello(msg);
			if(checkHeader >= 0) {
				System.out.println("TLSServerSpec::receiveHello:checkHello: " + strerr);
				disconnectWithSendAlert = checkHeader;
			}
		}

		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ServerHello);
		if(!isOutOfOrder) // i.e. msg.msg_type == 2
		{
			con.setRandom(msg.rnd, false);
			con.setAlgorithmParams(TLSCipherSuite.valueOf(msg.cipher_suites[0]));

			if(last_client_hello != null && msg.compareSessionID(last_client_hello.sessionID)) {
				DebugInfo.printMessage(3, "TLSServerSpec::receiveHello: session ID is resumed");
			}
			if(initial_handshake)
				initial_handshake = false;
			con.updateState(msg);
			mustSendEndOfEarlyData = con.getHandshakeParams().mustSendEndOfEarlyData;
			isClientPostAuth = con.getHandshakeParams().isClientPostAuth;
			isServerMustSendCertificate = HMCertificate.needServerCertificate(con.getHandshakeParams().keyExchangeMode);
		}
	}

	void receiveHelloRequest(HelloRequest msg) {
		isOutOfOrder = true;
		disconnectWithSendAlert = 10; // unexpected_message
	}

	void receiveEncryptedExtensions(EncryptedExtensions msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_EncryptedExtensions);
		con.updateState(msg);
	}

	void receiveClientKeyExchange(ClientKeyExchange msg) {
		isOutOfOrder = true;
		disconnectWithSendAlert = 10; // unexpected_message
	}

	void receiveEndOfEarlyData(EndOfEarlyData msg) {
		isOutOfOrder = true;
		disconnectWithSendAlert = 10; // unexpected_message
	}

	void receiveCertificateVerify(CertificateVerify msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		} else {
			disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg, con);
			if(disconnectWithSendAlert >= 0) {
				DebugInfo.printMessage(3, "===Error! CertificateVerify::verifySignature: failed");
				disconnectWithSendAlert = -1;
			}
		}
		
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_ServerCertificateVerify);
	}

	void receiveCertificate(HMCertificate msg, TLSConnection con) throws Exception {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);
		
		try {
			CertificateFactory factory = CertificateFactory.getInstance("X509");
			serverCert = (X509Certificate)factory.generateCertificate( new ByteArrayInputStream(msg.getFirstCertificate().getData()) );
			con.getHandshakeParams().remoteCert = (X509Certificate)factory.generateCertificate( new ByteArrayInputStream(msg.getFirstCertificate().getData()) );
			if(serverCert.getVersion() != 3) {  //  x509v3
				DebugInfo.printMessage(1,"===Error! receiveCertificate: recieved certificate x509 version failed: get " + serverCert.getVersion() + ", need 3(v3)");
				disconnectWithSendAlert = 43;						
			}
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! receiveCertificate: create certificate failed: " + err);
			serverCert = null;
			disconnectWithSendAlert = 50;
		}
		
		previousState = currentState;
		if(con.localRole == ConnectionRole.CLIENT)
			setCurrentState(TLSScenarioState.AFTER_ServerCertificate);
		else
			setCurrentState(TLSScenarioState.AFTER_ClientCertificate);
		
		con.updateState(msg);
	}

	void receiveServerKeyExchange (ServerKeyExchange msg, TLSSecurityParameters params) {
		isOutOfOrder = true;
		disconnectWithSendAlert = 10; // unexpected_message
	}

	void receiveCertificateRequest (CertificateRequest msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);

		isGetCertificateRequest = true;
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_CertificateRequest);
		
		con.updateState(msg);
	}

	void receiveServerHelloDone (ServerHelloDone msg) {
		isOutOfOrder = true;
		disconnectWithSendAlert = 10; // unexpected_message
	}

	void receiveFinished (Finished msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg, con, curMsgDir);
		
		previousState = currentState;
		if(con.localRole == ConnectionRole.CLIENT) {
			setCurrentState(TLSScenarioState.AFTER_ServerFinished);
			afterServerFinished = true;
		} else
			setCurrentState(TLSScenarioState.AFTER_ClientFinished);
		
		con.updateState(msg, curMsgDir);
		
		if((disconnectWithSendAlert < 0) && (currentState == TLSScenarioState.AFTER_ClientFinished))
			isHandshakeCompleted = true;

	}

	void receiveNewSessionTicket(NewSessionTicket msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);

		sessionTicketsList.add(msg);
		Date dt = new Date();
		byte[] newPSK = null;
		TLSSecurityParameters tlsParams = con.getSecurityParams();
		try {
			newPSK = TLSKeyDerivation.TLS_HKDF_ExpandLabel(tlsParams.mac_alg, tlsParams.resumption_master_secret, "resumption", msg.getTicketNonce());
		} catch (Exception e) {
    		DebugInfo.printMessage(1,"===Error! receiveNewSessionTicket: PSK create failed");
		}
		sessionTicketsPSK.add(new TicketPSK(dt.getTime(), newPSK));
		
		previousState = currentState;
		setCurrentState(TLSScenarioState.AFTER_NewSessionTicket);
	}

	void receiveKeyUpdate(KeyUpdate msg, TLSConnection con) {
		if(!CheckMessageOrder.checkOrder(msg.msg_type, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);
		
		previousState = currentState;
		if(con.localRole == ConnectionRole.CLIENT)
			setCurrentState(TLSScenarioState.AFTER_ServerKeyUpdate);
		else
			setCurrentState(TLSScenarioState.AFTER_ClientKeyUpdate);
		
		if(waitFatalAlert < 0)
			con.updateState(msg, curMsgDir);
	}

	void receiveChangeCipher (ChangeCipher msg) {
		if(!CheckMessageOrder.checkOrder(msg, this, curMsgDir)) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
		disconnectWithSendAlert = CheckMessageHeader.checkHeader(msg);
	}

	void receiveAlert (Alert msg) {
		if(waitFatalAlert > 0) {
			waitFatalAlert = -1;
			disconnectWithSendAlert = 0;
			previousState = currentState;
			setCurrentState(TLSScenarioState.DISCONNECTED);
			return;
		}
			
		checkHeader = checkAlert(msg);
		if(checkHeader >= 0) {
			System.out.println("TLSServerSpec::checkAlert: " + strerr);
			disconnectWithSendAlert = checkHeader;
		}
		else {
			if(msg.description == AlertType.CLOSE_NOTIFY.getValue()) {
					if(currentState == TLSScenarioState.AFTER_ServerCloseNotify) {
						isOutOfOrder = true;
						disconnectWithSendAlert = 10;							
					} else if(currentState == TLSScenarioState.AFTER_ClientCloseNotify) {
						previousState = currentState;
						setCurrentState(TLSScenarioState.DISCONNECTED);
					} else {
						previousState = currentState;
						setCurrentState(TLSScenarioState.AFTER_ServerCloseNotify);
					}
				}
			else {
				disconnectWithSendAlert = 0;
				previousState = currentState;
				setCurrentState(TLSScenarioState.DISCONNECTED);
				if(msg.description != waitFatalAlert) {
					DebugInfo.printMessage(1,"===Warning! receiveAlert: wait alert {2," + waitFatalAlert + "}, but received " + msg.toStringShort());
					isOutOfOrder = true;
				}
				waitFatalAlert = -1;
			}
		}
	}

	void receiveApplicationData (ApplicationData msg) {
		try {
			PrintWriter wrt = new PrintWriter(new OutputStreamWriter(new FileOutputStream("responseApplicationData.txt",true)));
			DebugInfo.printMessage(1,"TLSServerSpec::receive: write application data to file");
			wrt.print(new String(msg.data));
			wrt.close();
		} catch (FileNotFoundException ignored) {}
			
		if(  !isHandshakeCompleted
		  || currentState == TLSScenarioState.AFTER_ServerCloseNotify) {
			isOutOfOrder = true;
			disconnectWithSendAlert = 10; // unexpected_message
		}
	}


	int checkHello(Hello msg) {
		int alert = -1, i, ind;

		if(msg.msg_type == 2 && !msg.version.isTLS12())
					alert = 70;
		else if(msg.rnd.length != 32 
			||  msg.sessionID.length > 32 
			||  msg.cipher_suites.length == 0 
			||  (msg.msg_type == 2 && msg.cipher_suites.length != 1)
			||  msg.compression_methods.length != 1 
			||  msg.compression_methods[0] != 0) 
					alert = 47;
		else if( ( TLSCipherSuite.UNKNOWN == TLSCipherSuite.valueOf(msg.cipher_suites[0]) 
				   || TLSCipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV == TLSCipherSuite.valueOf(msg.cipher_suites[0]))
				 && msg.cipher_suites.length == 1)
			alert = 40;
		else if(TLSExtensionsList.checkDuplicateExtensions(msg.extensions))
				alert = 50;
		else if(msg.msg_type == 2) {
			if(last_client_hello == null)
				alert = 10;
			else {
				if(ConvertTypes.ntoh(last_client_hello.rnd) > ConvertTypes.ntoh(msg.rnd)) 
					// uint32 gmt_unix_time of Random structure
					System.out.println("===Warning! TLSServerSpec::checkHello: ServerHello.time < ClientHello.time");

				for(i=0; i<last_client_hello.cipher_suites.length; i++)
					if(msg.cipher_suites[0] == last_client_hello.cipher_suites[i])
						break;
				if(i == last_client_hello.cipher_suites.length) 
					return 47;
				
				TLSExtension curext;
				for(i=0; i<msg.extensions.size(); i++) {
					curext = (TLSExtension)msg.extensions.get(i);
					if(curext.type == (short)0xFF01)
						continue;
					ind = TLSExtensionsList.indexOfExtension(last_client_hello.extensions, curext.type);
					if(((ind == -1)&&(curext.type != 44)) || curext.type == 13) {
						alert = 110;
						break;
					}
				}
			}
		}

		DebugInfo.printMessage(5, "TLSServerSpec::checkHello: return " + alert);
		return alert;
	}

	int checkAlert(Alert msg) {
		return (msg.level == 1 || msg.level == 2) ? -1 : 10;
	}

	void idlingDecrypt(ArrayList<RecordLayer> msglst, int ind, TLSConnection con) {
		RecordLayer msgrl;
		
		DebugInfo.printMessage(3,"TLSServerSpec::idlingDecrypt, message index " + ind);

		if(msglst == null  ||  con == null) return;

		for(int i=ind; i < msglst.size(); i++) {
			msgrl = msglst.get(i);
			if(-1 != msgrl.decryptAndParseTLS(con))
				return;
			
			DebugInfo.printMessage(6,"TLSServerSpec::idlingDecrypt: increment Sequence Number");
			con.incrementReceiverSequenceNumber();

			for(Object msg : msgrl.msglist) {

				if(msg instanceof HandshakeMessage) {
					if(msg instanceof Hello) {
						Hello hl_msg = (Hello)msg;
						if(hl_msg.msg_type != 2)
							continue;
						if(last_client_hello != null && hl_msg.compareSessionID(last_client_hello.sessionID))
							isSessionResume = true;
						con.setRandom(hl_msg.rnd, false);
						con.setAlgorithmParams(TLSCipherSuite.valueOf(hl_msg.cipher_suites[0]));
					}
					else if(msg instanceof HMCertificate) {
						HMCertificate cert_msg = (HMCertificate)msg;
						try {
							CertificateFactory factory = CertificateFactory.getInstance("X509");
							serverCert = (X509Certificate)factory.generateCertificate( new ByteArrayInputStream(cert_msg.getFirstCertificate().getData()) );
						}
						catch (Exception err) {
							DebugInfo.printMessage(1,"===Error! idlingDecrypt: create certificate failed: " + err);
							serverCert = null;
						}
					}
				}
				else if(msg instanceof ChangeCipher) {
					setCurrentState(TLSScenarioState.AFTER_ServerChangeCipherSpec);
				}
			}
		}
			
		DebugInfo.printMessage(3,"TLSServerSpec::idlingDecrypt return");
		return;
	}

	public byte[] DisconnectWithAlert(TLSConnection current_connection) {
		byte[] msg = null;
		try {
			isOutOfOrder = false;
			if(disconnectWithSendAlert >= 0  &&  currentState != TLSScenarioState.DISCONNECTED) {
				System.out.println("\nSend closing alert");
				Alert al = new Alert(2,disconnectWithSendAlert);
				if(disconnectWithSendAlert == 0)
					al.level = 1;
				RecordLayer tlsmsg = new RecordLayer(al);
				tlsmsg.setContent(al,null,current_connection);
				disconnectWithSendAlert = -1;
				if(sock != null) {
					send(tlsmsg,current_connection);
				}
			}
		} catch (Exception ignored) {				
		} finally {
			System.out.println("Closing session");
			disconnect();	
		}
		return msg;
	}

}
