package ru.ispras.telecomtesk.tlsv13.TLSspecification;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.ChangeCipher;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.AlertType;
import ru.ispras.telecomtesk.tlsv13.TLSState.MessageDirection;
import ru.ispras.telecomtesk.tlsv13.TLSState.TLSConnection;

public class CheckMessageHeader
{
    public static int checkHeader(EncryptedExtensions msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader:EncryptedExtensions");
		// The client MUST check EncryptedExtensions for the
		//   presence of any forbidden extensions and if any are found MUST abort
		//   the handshake with an "illegal_parameter" alert.

		DebugInfo.printMessage(5,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(CertificateRequest msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader: CertificateRequest");
		
		//

		DebugInfo.printMessage(5,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(HMCertificate msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader: HMCertificate");
		
		//

		DebugInfo.printMessage(5,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(CertificateVerify msg, TLSConnection con) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader: CertificateVerify");
		try {
			if(!msg.verifySignature(con, msg.getSignature()))
				res = AlertType.DECRYPT_ERROR.getValue(); 
		} catch (Exception err) {
			err.printStackTrace();
		}

		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(Finished msg, TLSConnection con, MessageDirection msgDir) {
    	int res = -1;
    	byte[] data = null;

		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader: Finished");

		try {
			data = msg.calcVerifyData(con, con.getHandshakeTrafficSecret(msgDir));
		} catch (Exception err) {
			DebugInfo.printMessage(4,"===Error! CheckMessageHeader::checkHeader: Finished calcVerifyData failed");
			err.printStackTrace();
			return AlertType.DECRYPT_ERROR.getValue(); 
		}
		
		DebugInfo.printDump(5,msg.verify_data,"Signature get: ");
		DebugInfo.printDump(5,data,"Signature calculated: ");

		if((msg.verify_data == null) || 
			(msg.verify_data.length != data.length)) {
			DebugInfo.printMessage(1,"===Error! Finished verifySignature: compare data length failed: " + msg.verify_data.length + " " + data.length);
			res = AlertType.DECRYPT_ERROR.getValue(); 
		}
		else {
			for(int i=0; i < data.length; i++)
				if(msg.verify_data[i] != data[i]) {
					DebugInfo.printMessage(1,"===Error! Finished verifySignature: compare data failed: ");
					DebugInfo.printDump(1,msg.verify_data,"   verifySignature: get: ");
					DebugInfo.printDump(1,data,"   verifySignature: need: ");
					res = AlertType.DECRYPT_ERROR.getValue(); 
					break;
				}
		}
		
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(KeyUpdate msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader:KeyUpdate");
		if((msg.requestUpdate != 0) && (msg.requestUpdate != 1))
			res = AlertType.ILLEGAL_PARAMETER.getValue();
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(NewSessionTicket msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader:NewSessionTicket");
		DebugInfo.printMessage(5,"CheckMessageHeader::checkHeader return " + res);
    	return res;
    }
    	
    public static int checkHeader(ChangeCipher msg) {
    	int res = -1;
		DebugInfo.printMessage(4,"CheckMessageHeader::checkHeader: ChangeCipher");
		if (msg.flag != 1) res = AlertType.UNEXPECTED_MESSAGE.getValue(); 
    	return res;
    }
    	
}
