package ru.ispras.telecomtesk.tlsv13.TestsServer.CheckTlsConnection;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.io.IOException;
import java.math.BigInteger;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;

import sun.security.ec.ECParameters;
import java.util.Date;
import java.util.Random;

import javax.crypto.KeyAgreement;
import javax.crypto.interfaces.DHPrivateKey;
import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;
import javax.crypto.spec.DHPublicKeySpec;
import javax.xml.bind.DatatypeConverter;

import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.TLSserverSpecification.TLSServerSpec;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSSocket.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.LoginInfo;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

class CreateClientHello {
	protected String targetServer;
	protected int targetPort;
	protected String socketTimeout;
	public LoginInfo LI = new LoginInfo();
	public TLSConnection current_connection;
	public TLSServerSpec target = new TLSServerSpec();
	public InetAddress hostIP;
	public String applicationDataString = null;

	public CreateClientHello(LoginInfo scenProp) {
		LI = scenProp;
		DebugInfo.debugLevel = 4;
		DebugInfo.maxDumpSize = 800;

		targetServer = scenProp.getExistingValue("server");
		if (targetServer == null) {
			throw new IllegalArgumentException("Server not specified");
		}

		String targetPortString = scenProp.getExistingValue("portTLS");
		if (targetPortString == null) {
			throw new IllegalArgumentException("Server port not specified");
		}
		targetPort = Integer.parseInt(targetPortString);

		applicationDataString = scenProp.getValue("applicationData", "Test message");

		socketTimeout = scenProp.getValue("timeout");

		TLSConnectionSelector sel = new TLSConnectionSelector(null,null,0,0);
		current_connection = new TLSConnection(sel,ConnectionRole.CLIENT);
		// by default current_connection.getPendingState().getParams().isServer = false; 
	
		try {
			hostIP = InetAddress.getLocalHost();
		} catch(UnknownHostException err) {
			throw new RuntimeException( "Failed to initialize scenario",err);
		}

		target.sock = new TLSSocket(targetServer, targetPort, Integer.parseInt(socketTimeout));
	}

	public void ClientHello() {
		DebugInfo.printMessageWithSeparator(1,"scenario ClientHello start");
		ArrayList<RecordLayer> response;
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue(),
					  TLSCipherSuite.TLS_AES_128_CCM_SHA256.getValue()
					 };
		Hello msg;

		ProtocolVersion svdata = new ProtocolVersion(0x03,0x04);
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);

		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256, SignatureScheme.ECDSA_SECP256R1_SHA256};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);
		SignatureAndHashAlgorithmExt hl_ext_sah2 = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);


	    BigInteger g = new BigInteger("2", 16);
	    DHParameterSpec dhsp = new DHParameterSpec(TLS13SpecValues.p2048, g);
		ECGenParameterSpec ecsp = new ECGenParameterSpec("secp256r1");


		KeyPair keyPair = null;
		try {
//	    KeyPairGenerator keyGen = KeyPairGenerator.getInstance("DiffieHellman");
 //       keyGen.initialize(dhsp);
	    KeyPairGenerator keyGen = KeyPairGenerator.getInstance("EC");
	    keyGen.initialize(new ECGenParameterSpec("secp256r1"), new SecureRandom());
	    keyPair = keyGen.generateKeyPair();
		} catch (Exception e) {
			
		}

        PrivateKey priv = keyPair.getPrivate();
        PublicKey pub = keyPair.getPublic();
//        DHPublicKey clientPublicKey = (DHPublicKey)pub; 
//        byte[] DH_Yc = clientPublicKey.getY().toByteArray();
        ECPublicKey clientPublicKey = (ECPublicKey)pub; 
        byte[] DH_Xc = clientPublicKey.getW().getAffineX().toByteArray();
        byte[] DH_Yc = clientPublicKey.getW().getAffineY().toByteArray();
        if(DH_Xc[0] == 0x00) 
        	DH_Xc = Arrays.copyOfRange(DH_Xc, 1, DH_Xc.length);
        if(DH_Yc[0] == 0x00) 
        	DH_Yc = Arrays.copyOfRange(DH_Yc, 1, DH_Yc.length);
        
		DebugInfo.printDumpWithLength(1, DH_Xc, "PublicKey: ");
		DebugInfo.printDumpWithLength(1, DH_Yc, "PublicKey: ");

		
		
	    byte[] rX = DatatypeConverter.parseHexBinary(
	    		"E1ADEA6B9CD5BB7FE239B52900390915" +
	    	    "2FD1EB264F42A36B02F17DB4350F8CC1");
	    byte[] rY = DatatypeConverter.parseHexBinary(
	    		"C52D7BC870CD9A3B55165F55203E3EE7" +
	    	    "2EBDEB28CBF656856449977726D6B4E6");
	    BigInteger rXb = new BigInteger(
	    		"E1ADEA6B9CD5BB7FE239B52900390915" +
	    	    "2FD1EB264F42A36B02F17DB4350F8CC1", 16);
	    BigInteger rYb = new BigInteger(
	    		"C52D7BC870CD9A3B55165F55203E3EE7" +
	    	    "2EBDEB28CBF656856449977726D6B4E6", 16);
	    BigInteger rXb2 = new BigInteger(
	    		"3949795390bad7c41d80a942785cc60b" +
	    	    "ed83f538146cf8e73280fb170cb3761b", 16);
	    BigInteger rYb2 = new BigInteger(
	    		"ddb7563f64d9e253ed54d5e733638f31" +
	    	    "956dce4806d8e6a208ee4df9d07e09cd", 16);

	    String st1 = DatatypeConverter.printHexBinary(rX);
	    String st2 = DatatypeConverter.printHexBinary(rY);
		DebugInfo.printMessage(1, "----");
		DebugInfo.printMessage(1, st1);
		DebugInfo.printMessage(1, st2);
	    	    
	    ECPoint rP = new ECPoint(rXb2,rYb2);

	    try {
		ECPublicKeySpec dhKeySpec = new ECPublicKeySpec(rP, clientPublicKey.getParams());
		KeyFactory localKeyFactory = KeyFactory.getInstance("EC");
		ECPublicKey remoteDHPublicKey = (ECPublicKey) localKeyFactory.generatePublic(dhKeySpec);

		KeyAgreement keyAgree = KeyAgreement.getInstance("ECDH");
		keyAgree.init(priv);
		keyAgree.doPhase(remoteDHPublicKey, true);
		byte[] DHsecret = keyAgree.generateSecret();
		DebugInfo.printDumpWithLength(1, DHsecret, "DHsecret: ");
	    }
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! KeyAgreement Exception: " + err);
			return;
		}

	    try {
	    	KeyShareEntry kse = new KeyShareEntry(NamedGroup.SECP256R1, pub);
	    }
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! KeyShareEntry Exception: " + err);
			return;
		}
//		KeyShareEntry kse = new KeyShareEntry(NamedGroup.FFDHE2048, DH_Yc);
//		TLSExtension hl_ext_ks = new KeyShareClientHello(kse);
//		TLSExtension hl_ext_ks = new KeyShareClientHello();
/*
		byte[] dataext = {1,2,3,4,5};
		TLSExtension hl_ext_un = new TLSExtension(1020,dataext);
		TLSExtension hl_ext_un2 = new TLSExtension(1020,dataext);

		// create Hello
		DebugInfo.printMessage(1,"create Hello");
		msg = new Hello(cs);
		msg.addExtension(hl_ext_sv);
		msg.addExtension(hl_ext_sah);
		msg.addExtension(hl_ext_sg);
		msg.addExtension(hl_ext_ks);
		
		DebugInfo.printMessage(1,"print Hello");
		msg.print();
		DebugInfo.printMessage(1,"toByteArray Hello");
		msg.toByteArray();
		DebugInfo.printMessage(1,"end Hello");
		
		RecordLayer tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		tlsmsg.msglist.add(msg);
		
		try {
			tlsmsg.setContent(msg,null,current_connection);
			DebugInfo.printDump(5, msg.net_msg, "ClientHello: ");
			response = target.send(tlsmsg,current_connection);
			if(null == response  ||  target.disconnectWithSendAlert != -1) {
				DebugInfo.printMessage(1, "ClientHello send failed");
				return;
			}
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! scenario ClientHello Exception: " + err);
			target.disconnect();	
			DebugInfo.printMessage(1, "Exception occured");
			return;
		}
		
		finally {
			target.DisconnectWithAlert(current_connection);
		}
*/
		return;
	}
	
	

	public static void main(String[] args) {
		Date dt1 = new Date();
		System.out.println(dt1.toLocaleString());
//		String fileName = "connectInfo_mail.txt";
//		String fileName = "connectInfo_http.txt";
		String fileName = "connectInfo_java.txt";
		if (args.length > 0 )
			fileName = args[0];
		start(fileName);
		Date dt2 = new Date();
		System.out.println(dt2.toLocaleString());
	}
	
	
	public static void start(String fileName) {
		
		LoginInfo LI = new LoginInfo();
		LI.readFromFile(fileName);
		
		CreateClientHello scen = new CreateClientHello(LI);
		scen.ClientHello();
	}
}