package ru.ispras.telecomtesk.tlsv13.TestsServer.CheckTlsConnection;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.lang.invoke.MethodHandles;
import java.math.BigInteger;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;

import sun.security.ec.ECParameters;

import javax.crypto.KeyAgreement;
import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;
import javax.xml.bind.DatatypeConverter;

import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.TLSserverSpecification.TLSServerSpec;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSSocket.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.LoginInfo;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

class TestAdditionalCheck {
	protected String targetServer;
	protected int targetPort;
	protected String socketTimeout;
	protected boolean testSuitEnd = false;
	protected Hello msgGl;
	public LoginInfo LI = new LoginInfo();
	public TLSConnection current_connection;
	public TLSServerSpec target = new TLSServerSpec();
	public InetAddress hostIP;
	public String applicationDataString = "Test message";
	public static long testCounter = 0;
	public static PrintWriter logFile;

	public TestAdditionalCheck(LoginInfo scenProp) {
		LI = scenProp;
		DebugInfo.debugLevel = 4;
		DebugInfo.maxDumpSize = 600;

		targetServer = scenProp.getExistingValue("server");
		if (targetServer == null) {
			throw new IllegalArgumentException("Server not specified");
		}

		String targetPortString = scenProp.getExistingValue("portTLS");
		if (targetPortString == null) {
			throw new IllegalArgumentException("Server port not specified");
		}
		targetPort = Integer.parseInt(targetPortString);

		applicationDataString = scenProp.getValue("applicationData", "Test message");

		socketTimeout = scenProp.getValue("timeout");

		TLSConnectionSelector sel = new TLSConnectionSelector(null,null,0,0);
		current_connection = new TLSConnection(sel,ConnectionRole.CLIENT);
		// by default current_connection.getPendingState().getParams().isServer = false; 
	
		try {
			hostIP = InetAddress.getLocalHost();
		} catch(UnknownHostException err) {
			throw new RuntimeException( "Failed to initialize scenario",err);
		}

		target.sock = new TLSSocket(targetServer, targetPort, Integer.parseInt(socketTimeout));
	}

	public void connect() {
		DebugInfo.printMessageWithSeparator(1,"scenario connect start");
		try {
			target.connect(current_connection);
		}
		catch (IOException err) {
			System.out.println("scenario Connect failed" + err);
		}
	}

	public void  ClientHello_pre() {
		String scenName = " ClientHello ";
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario start," + scenName);
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue(),
				TLSCipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA.getValue(),
				TLSCipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA256.getValue()
				 };
		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256
				, SignatureScheme.ECDSA_SECP256R1_SHA256
				, SignatureScheme.RSA_PSS_RSAE_SHA256
				};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.FFDHEtest, NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);

		NamedGroup[] sendgrp = {NamedGroup.FFDHEtest};
		current_connection.hskParams.generateClientKeyPairList(sendgrp);

		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04),new ProtocolVersion(0x03,0x03)};
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);
		TLSExtension hl_ext_ks = new KeyShareClientHello(current_connection.hskParams.clientKeyShareList);
		DebugInfo.printMessage(1,"create Hello");
		byte[] dataId = {1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,7,8};
		msgGl = new Hello(dataId,cs);
		msgGl.addExtension(hl_ext_sv);
		msgGl.addExtension(hl_ext_sah);
		msgGl.addExtension(hl_ext_sg);
		msgGl.addExtension(hl_ext_ks);

		Exception err = msgSetContentAndSend(msgGl,null,scenName);
		if(err != null) {
			throw new RuntimeException( "scenario" + scenName + ": msgSetContentAndSend failed ",err);
		}
		return;
	}

	public void ClientHello() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario ClientHello start");
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue()
//					TLSCipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA.getValue(),
//					TLSCipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA256.getValue()
					 };
		
		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256
				, SignatureScheme.ECDSA_SECP256R1_SHA256
				, SignatureScheme.RSA_PSS_RSAE_SHA256
				};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.FFDHEtest, NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);

//		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04)};
		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04),new ProtocolVersion(0x03,0x03)};
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);

//		NamedGroup[] sendgrp = {NamedGroup.FFDHE2048};
		NamedGroup[] sendgrp = {NamedGroup.SECP256R1};
//		NamedGroup[] sendgrp = {NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		current_connection.hskParams.generateClientKeyPairList(sendgrp);
		TLSExtension hl_ext_ks = new KeyShareClientHello(current_connection.hskParams.clientKeyShareList);
		

		String snStr1 = "test.example.com";
		String snStr2 = "noname.noname.com";
		ServerNameExt snExt = new ServerNameExt(snStr1);
		snExt.addServerName(snStr2);
		
		// create Hello
		DebugInfo.printMessage(1,"create Hello");
		byte[] dataId = {1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,7,8};
		Hello msg;
		if((msgGl != null) && (msgGl.rnd != null))
			msg = new Hello(msgGl.rnd, dataId,cs);
		else
			msg = new Hello(dataId,cs);
		
		msg.addExtension(hl_ext_sv);
		msg.addExtension(snExt);
		msg.addExtension(hl_ext_sah);
		msg.addExtension(hl_ext_sg);
		if(target.needCookie) {
//			byte[] dtc = new byte[current_connection.hskParams.remoteCookie.length];
//			System.arraycopy(current_connection.hskParams.remoteCookie, 0, dtc, 0, dtc.length);
//			dtc[10] += 1;
//			CookieExt hl_ext_ck = new CookieExt(dtc);
			
			CookieExt hl_ext_ck = new CookieExt(current_connection.hskParams.remoteCookie);
//			CookieExt hl_ext_ck = new CookieExt(null);
//			TLSExtension hl_ext_ck = new TLSExtension(44,new byte[0]);
			msg.addExtension(hl_ext_ck);
		}
		msg.addExtension(hl_ext_ks);
		
		DebugInfo.printMessage(1,"print Hello");
		msg.print();
		
		Exception err = msgSetContentAndSend(msg,null,"ClientHello");
		if(err != null) {
			throw new RuntimeException( "scenario ClientHello: msgSetContentAndSend failed",err);
		}
		
		return;
	}
	
	public void ClientHello_res() {
		if(target.getCurrentState() != TLSScenarioState.CONNECTED)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario ClientHello start");
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue()
					 };
		
		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256
				, SignatureScheme.ECDSA_SECP256R1_SHA256
				, SignatureScheme.RSA_PSS_RSAE_SHA256
				};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.SECP256R1, NamedGroup.FFDHE2048};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);

		NamedGroup[] sendgrp = {NamedGroup.FFDHE2048};
//		NamedGroup[] sendgrp = {NamedGroup.SECP256R1};
		current_connection.hskParams.generateClientKeyPairList(sendgrp);

		TLSExtension hl_ext_ks = new KeyShareClientHello(current_connection.hskParams.clientKeyShareList);

		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04)};
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);
		
		byte[] pksModes = {1};
		PSKexchangeModesExt pksModesExt = new PSKexchangeModesExt(pksModes);

		NewSessionTicket tc1 = target.sessionTicketsList.get(0);
		TicketPSK tk = target.sessionTicketsPSK.get(0);
		Date dt = new Date();
		int obAge = (int)(dt.getTime() - tk.ticketRecvTime + tc1.ticketAgeAdd);
		PskIdentity pskid = new PskIdentity(obAge, tc1.getTicket());

		MACAlgorithm malg = TLSCipherSuite.valueOf(cs[0]).getMacAlgorithm();
		byte[] binder = new byte[malg.getHashLength()];
		
		PreSharedKeyClientHello pskCH = new PreSharedKeyClientHello(pskid, binder);

		// create Hello
		DebugInfo.printMessage(1,"create Hello");
		byte[] dataId = {1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,7,8};
		Hello msg = new Hello(dataId,cs);
		msg.addExtension(hl_ext_sv);
		msg.addExtension(hl_ext_sah);
		msg.addExtension(hl_ext_sg);
		msg.addExtension(hl_ext_ks);
		msg.addExtension(pksModesExt);
		msg.addExtension(pskCH);

		msg.toByteArray();
		DebugInfo.printMessage(1,"print Hello");
		msg.print();
		DebugInfo.printDump(1,msg.net_msg,"");

		int bsz = 2 + pskCH.getBindersSize();
		byte[] msgHash = new byte[msg.net_msg.length - bsz];
		System.arraycopy(msg.net_msg, 0, msgHash, 0, msgHash.length);
		DebugInfo.printDump(1,msgHash,"msgHash");
		
		try {
			byte[] early_secret = TLSKeyDerivation.TLS_HKDF_Extract(malg, null, tk.psk);
			byte[] binder_key = TLSKeyDerivation.TLS_DeriveSecret(malg, early_secret, "res binder", null);
			byte[] hashHandshakeContext = TLSKeyDerivation.hash(malg, msgHash);
			byte[] macKey = TLSKeyDerivation.TLS_HKDF_ExpandLabel(malg, binder_key, "finished", null);
			binder = TLSKeyDerivation.mac(malg,macKey,hashHandshakeContext);
		} catch (Exception e) {
			DebugInfo.printMessage(1,"Error!!! calculate binder failed");
		}

		msg.removeExtension(pskCH);
		pskCH.binders.clear();
		pskCH.binders.add(binder);
		msg.addExtension(pskCH);
		
		
		msg.toByteArray();
		DebugInfo.printDump(1,msg.net_msg,"Hello net_msg");
		
		Exception err = msgSetContentAndSend(msg,null,"ClientHello");
		if(err != null) {
			throw new RuntimeException( "scenario ClientHello: msgSetContentAndSend failed",err);
		}
		
		return;
	}
	
	public void Finished() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario Finished start");
		Finished msg = null;
		try {
			msg = new Finished(current_connection, MessageDirection.getMessageDirection(current_connection.localRole, TLSDirection.SEND)); 
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Warning! create new Finished message failed");
		}
		
		Exception err = msgSetContentAndSend(msg,null,"Finished");
		if(err != null) {
			throw new RuntimeException( "scenario Finished: msgSetContentAndSend failed",err);
		}
		
		return;
	}
	
	public void ChangeCipher() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario ChangeCipher start");
		ChangeCipher msg = new ChangeCipher();
		
		RecordLayer tlsmsg = new RecordLayer(TLSContentType.CHANGE_CIPHER_SPEC, msg);
		tlsmsg.initContent(1);
		tlsmsg.content[0] = msg.toByteArray();
		target.mayNoResponse = true;

		Exception err = msgSend(tlsmsg, "ChangeCipher");
		if(err != null) {
			throw new RuntimeException( "scenario ChangeCipher: msgSend failed",err);
		}
		
		return;
	}
	
	public void ApplicationData() {
		if(!target.getCurrentState().connected)
			return;
		
		DebugInfo.printMessageWithSeparator(1,"scenario ApplicationData start");
		System.out.println(applicationDataString);
		ApplicationData msg = new ApplicationData(applicationDataString.getBytes());

		RecordLayer rlmsg = new RecordLayer(msg);
		try {
			rlmsg.setContent(msg,null,current_connection);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
			target.disconnect();
			return;
		}
		Exception err = msgSendWithoutAlert(rlmsg, "ApplicationData");
//		Exception err = msgSetContentAndSend(msg,null,"ApplicationData");
		if(err != null) {
			throw new RuntimeException( "scenario ApplicationData: msgSetContentAndSend failed",err);
		}

		if(testSuitEnd) {
		target.disconnectWithSendAlert = 0;
		target.DisconnectWithAlert(current_connection);
		}

		return;
	}
	
	
	public void KeyUpdate() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario KeyUpdate start");
		KeyUpdate msg = new KeyUpdate(1);
		
		Exception err = msgSetContentAndSend(msg,null,"KeyUpdate");
		if(err != null) {
			throw new RuntimeException( "scenario KeyUpdate: msgSetContentAndSend failed",err);
		}
		
		testSuitEnd = true;
		return;
	}
	
	public void RecordLayerUndefinedTypeMessage() {
		String scenName = " RecordLayerUndefinedTypeMessage ";
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario start," + scenName);
		byte[] data = {12,34,56,83,23,45};
		HandshakeUndefinedType msg = new HandshakeUndefinedType(37,data);
		msg.toByteArray();
		RecordLayer tlsmsg = new RecordLayer(msg);
		
		try {
			tlsmsg.tls_type = TLSContentType.TESTTYPEONLY;
			tlsmsg.setContent(msg.net_msg,null,current_connection);
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! scenario setContent Exception: " + err);
			target.disconnect();	
			return;
		}

		Exception err = msgSend(tlsmsg, scenName);
		if(err != null) {
			throw new RuntimeException( "scenario" + scenName + ": msgSetContentAndSend failed ",err);
		}
		return;
	}

	public void EmptyDataAlert() {
		String scenName = " EmptyDataAlert ";
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario start," + scenName);
		byte[] data = new byte[0];
		RecordLayer tlsmsg = new RecordLayer(data);

		try {
			tlsmsg.tls_type = TLSContentType.ALERT;
			tlsmsg.setContent(data,null,current_connection);
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! scenario setContent Exception: " + err);
			target.disconnect();	
			return;
		}
		

		target.waitFatalAlert = 20;
		Exception err = msgSendWithoutAlert(tlsmsg, scenName);
		if(err != null) {
			throw new RuntimeException( "scenario" + scenName + ": msgSetContentAndSend failed ",err);
		}
		return;
	}
	

	public void EmptyDataHandshake() {
		String scenName = " EmptyDataHandshake ";
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario start," + scenName);
		byte[] data = new byte[0];
		RecordLayer tlsmsg = new RecordLayer(data);

		try {
			tlsmsg.tls_type = TLSContentType.HANDSHAKE;
			tlsmsg.setContent(data,null,current_connection);
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! scenario setContent Exception: " + err);
			target.disconnect();	
			return;
		}

		target.waitFatalAlert = 20;
		Exception err = msgSendWithoutAlert(tlsmsg, scenName);
		if(err != null) {
			throw new RuntimeException( "scenario" + scenName + ": msgSetContentAndSend failed ",err);
		}
		
		return;
	}
	

	public Exception msgSetContent(Object msg, byte[] padding, String scenName) {
		return msgSetContent(null, new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContent(TLSContentType tp, Object msg, byte[] padding, String scenName) {
		return msgSetContent(tp, new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContent(RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		return msgSetContent(null, rlmsg, msg, padding, scenName);
	}
	
	public Exception msgSetContent(TLSContentType tp, RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			DebugInfo.printMessage(1," --- Object --- ");
			rlmsg.setContent(msg,padding,current_connection);
			if(tp != null) {
				rlmsg.tls_type = tp;
				rlmsg.opaque_tls_type = tp;
			}
			DebugInfo.printDump(1, rlmsg.toByteArray(), scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
//			target.disconnect();
			return res;
		}
		return res;
	}
	
	
	public Exception msgSetContent(TLSContentType tp, RecordLayer rlmsg, byte[] msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			DebugInfo.printMessage(1," --- byte[] --- ");
			rlmsg.setContent(msg,padding,current_connection);
			if(tp != null) {
				rlmsg.tls_type = tp;
				rlmsg.opaque_tls_type = tp;
			}
			DebugInfo.printDump(1, rlmsg.toByteArray(), scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
//			target.disconnect();
			return res;
		}
		return res;
	}
	
	
	public Exception msgSetContentAndSend(Object msg, byte[] padding, String scenName) {
		return msgSetContentAndSend(new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContentAndSend(RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			rlmsg.setContent(msg,padding,current_connection);
			DebugInfo.printDump(5, rlmsg.toByteArray(), scenName);
//			if(msg instanceof HandshakeMessage)
//				DebugInfo.printDump(5, ((HandshakeMessage)msg).net_msg, scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
			target.disconnect();
			return res;
		}

		res = msgSend(rlmsg, scenName);
		
		return res;
	}
	
	
	public Exception msgSend(RecordLayer rlmsg, String scenName) {
		return msgSend(rlmsg, scenName, false);
	}
	
	public Exception msgSendWithoutAlert(RecordLayer rlmsg, String scenName) {
		return msgSend(rlmsg, scenName, false);
	}
	
	public Exception msgSend(RecordLayer rlmsg, String scenName, boolean sendAlert) {
		Exception res = null;
		
		try {
			ArrayList<RecordLayer> response = target.send(rlmsg,current_connection);
			if(null == response) {
				DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
				DebugInfo.logtrace(logFile,"===Error! scenario send response null, may be decrypt failed");
				target.disconnectWithSendAlert = 80;
				target.DisconnectWithAlert(current_connection);
			}
			else if((target.disconnectWithSendAlert != -1) && (sendAlert)) {
				target.DisconnectWithAlert(current_connection);
			}
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile,"===Error! scenario send Exception: " + err);
			target.disconnect();	
			res = err;
		}
		
		return res;
	}
	
	
	public static void main(String[] args) {
		String fileName = "connectInfo_mail.txt";
//		String fileName = "connectInfo_http.txt";
//		String fileName = "connectInfo_java.txt";
		if (args.length > 0 )
			fileName = args[0];
		start(fileName);
	}
	
	
	public static void start(String fileName) {
		String currentClassName = MethodHandles.lookup().lookupClass().getSimpleName();
		try {
			logFile = new PrintWriter(new OutputStreamWriter(new FileOutputStream(currentClassName + "_errorlog.txt",true)));
			DebugInfo.printMessage(1,"open log file " + currentClassName + "_errorlog.txt");
		}
		catch (FileNotFoundException err) {
			DebugInfo.printMessage(1,"open log file Exception: " + err);
			return;
		}
		Date dt1 = new Date();
		DebugInfo.logtrace(logFile,"start test " + currentClassName + " " + dt1.toString());
		
		LoginInfo LI = new LoginInfo();
		LI.readFromFile(fileName);
		
		TestAdditionalCheck scen = new TestAdditionalCheck(LI);
		
		scen.connect();
//		scen.ClientHello_pre();
		scen.ClientHello();
		scen.ChangeCipher();
		scen.Finished();
//		scen.RecordLayerUndefinedTypeMessage();
//		scen.EmptyDataAlert();
//		scen.EmptyDataHandshake();
		scen.ApplicationData();
//		scen.KeyUpdate();
//		scen.connect();
//		scen.ClientHello_res();

		dt1 = new Date();
		DebugInfo.logtrace(logFile,"end test: " + dt1.toString() + " testCounter " + testCounter);
	}
}