package ru.ispras.telecomtesk.tlsv13.TestsServer.CheckTlsConnection;

import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.*;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.lang.invoke.MethodHandles;
import java.math.BigInteger;
import java.security.Key;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.ECPoint;

import sun.security.ec.ECParameters;

import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;
import javax.net.ServerSocketFactory;
import javax.net.ssl.SSLServerSocket;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocket;
import javax.xml.bind.DatatypeConverter;

import ru.ispras.telecomtesk.tlsv13.TLSState.*;
import ru.ispras.telecomtesk.tlsv13.TLSserverSpecification.TLSServerSpec;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSSocket.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.LoginInfo;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

class TestClientHelloRetry {
	protected String targetServer;
	protected int targetPort;
	protected String socketTimeout;
	protected boolean testSuitEnd = false;
	protected Hello msgGl;
	public LoginInfo LI = new LoginInfo();
	public TLSConnection current_connection;
	public TLSServerSpec target = new TLSServerSpec();
	public InetAddress hostIP;
	public String applicationDataString = "Test message";
	public static long testCounter = 0;
	public static PrintWriter logFile;

	public TestClientHelloRetry(LoginInfo scenProp) {
		LI = scenProp;
		DebugInfo.debugLevel = 4;
		DebugInfo.maxDumpSize = 800;

		targetServer = scenProp.getExistingValue("server");
		if (targetServer == null) {
			throw new IllegalArgumentException("Server not specified");
		}

		String targetPortString = scenProp.getExistingValue("portTLS");
		if (targetPortString == null) {
			throw new IllegalArgumentException("Server port not specified");
		}
		targetPort = Integer.parseInt(targetPortString);

		applicationDataString = scenProp.getValue("applicationData", "Test message");

		socketTimeout = scenProp.getValue("timeout");

		TLSConnectionSelector sel = new TLSConnectionSelector(null,null,0,0);
		current_connection = new TLSConnection(sel,ConnectionRole.CLIENT);
	
		try {
			hostIP = InetAddress.getLocalHost();
		} catch(UnknownHostException err) {
			throw new RuntimeException( "Failed to initialize scenario",err);
		}

		target.sock = new TLSSocket(targetServer, targetPort, Integer.parseInt(socketTimeout));
	}

	public void connect() {
		DebugInfo.printMessageWithSeparator(1,"scenario connect start");
		try {
			target.connect(current_connection);
		}
		catch (IOException err) {
			System.out.println("scenario Connect failed" + err);
		}
	}

	public void  ClientHello_pre() {
		String scenName = " ClientHello_pre ";
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario start," + scenName);
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue()
				, TLSCipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA.getValue()
				, TLSCipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA256.getValue()
				 };
		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256
				, SignatureScheme.ECDSA_SECP256R1_SHA256
				, SignatureScheme.RSA_PSS_RSAE_SHA256
				};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.FFDHEtest, NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);

		NamedGroup[] sendgrp = {NamedGroup.FFDHEtest};
		current_connection.hskParams.generateClientKeyPairList(sendgrp);

		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04),new ProtocolVersion(0x03,0x03)};
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);
		TLSExtension hl_ext_ks = new KeyShareClientHello(current_connection.hskParams.clientKeyShareList);
		DebugInfo.printMessage(1,"create Hello");
		byte[] dataId = {1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,7,8};
		msgGl = new Hello(dataId,cs);
		msgGl.addExtension(hl_ext_sv);
		msgGl.addExtension(hl_ext_sah);
		msgGl.addExtension(hl_ext_sg);
		msgGl.addExtension(hl_ext_ks);
		msgGl.print();

		Exception err = msgSetContentAndSend(msgGl,null,scenName);
		if(err != null) {
			throw new RuntimeException( "scenario" + scenName + ": msgSetContentAndSend failed ",err);
		}
		return;
	}

	public void ClientHello() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario ClientHello start");
		short[] cs = {TLSCipherSuite.TLS_AES_128_GCM_SHA256.getValue()
				, TLSCipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA.getValue()
				, TLSCipherSuite.TLS_RSA_WITH_AES_256_CBC_SHA256.getValue()
				 };
		
		SignatureScheme[] sah = {SignatureScheme.RSA_PKCS1_SHA256
				, SignatureScheme.ECDSA_SECP256R1_SHA256
				, SignatureScheme.RSA_PSS_RSAE_SHA256
				};
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sah);

		NamedGroup[] sg = {NamedGroup.FFDHEtest, NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		SupportedGroupsExt hl_ext_sg = new SupportedGroupsExt(sg);

		ProtocolVersion[] svdata = {new ProtocolVersion(0x03,0x04),new ProtocolVersion(0x03,0x03)};
		TLSExtension hl_ext_sv = new SupportedVersionsExt(svdata);

//		NamedGroup[] sendgrp = {NamedGroup.FFDHE2048};
		NamedGroup[] sendgrp = {NamedGroup.SECP256R1};
//		NamedGroup[] sendgrp = {NamedGroup.FFDHE2048, NamedGroup.SECP256R1};
		current_connection.hskParams.generateClientKeyPairList(sendgrp);
		TLSExtension hl_ext_ks = new KeyShareClientHello(current_connection.hskParams.clientKeyShareList);
		

		String snStr1 = "test.example.com";
		String snStr2 = "noname.noname.com";
		ServerNameExt snExt = new ServerNameExt(snStr1);
//		snExt.addServerName(snStr2);
		
		// create Hello
		DebugInfo.printMessage(1,"create Hello");
		byte[] dataId = {1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,1,2,3,4,5,6,7,8};
		Hello msg;
		if((msgGl != null) && (msgGl.rnd != null))
			msg = new Hello(msgGl.rnd, dataId,cs);
		else
			msg = new Hello(dataId,cs);
		
		msg.addExtension(hl_ext_sv);
//		msg.addExtension(snExt);
		msg.addExtension(hl_ext_sah);
		msg.addExtension(hl_ext_sg);
		if(target.needCookie) {
//			byte[] dtc = new byte[current_connection.hskParams.remoteCookie.length];
//			System.arraycopy(current_connection.hskParams.remoteCookie, 0, dtc, 0, dtc.length);
//			dtc[10] += 1;
//			CookieExt hl_ext_ck = new CookieExt(dtc);
			
			CookieExt hl_ext_ck = new CookieExt(current_connection.hskParams.remoteCookie);
//			CookieExt hl_ext_ck = new CookieExt(null);
//			CookieExt hl_ext_ck = new CookieExt(dataId);
//			TLSExtension hl_ext_ck = new TLSExtension(44,new byte[0]);
			msg.addExtension(hl_ext_ck);
		}
		msg.addExtension(hl_ext_ks);
		
		DebugInfo.printMessage(1,"print Hello");
		msg.print();
		
		Exception err = msgSetContentAndSend(msg,null,"ClientHello");
		if(err != null) {
			throw new RuntimeException( "scenario ClientHello: msgSetContentAndSend failed",err);
		}
		
		return;
	}
	
	public void Finished() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario Finished start");
		Finished msg = null;
		try {
			msg = new Finished(current_connection, MessageDirection.getMessageDirection(current_connection.localRole, TLSDirection.SEND)); 
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Warning! create new Finished message failed");
		}
		
		Exception err = msgSetContentAndSend(msg,null,"Finished");
		if(err != null) {
			throw new RuntimeException( "scenario Finished: msgSetContentAndSend failed",err);
		}
		
		return;
	}
	
	public void ChangeCipher() {
		if(!target.getCurrentState().connected)
			return;
		DebugInfo.printMessageWithSeparator(1,"scenario ChangeCipher start");
		ChangeCipher msg = new ChangeCipher();
		
		RecordLayer tlsmsg = new RecordLayer(TLSContentType.CHANGE_CIPHER_SPEC, msg);
		tlsmsg.initContent(1);
		tlsmsg.content[0] = msg.toByteArray();
		target.mayNoResponse = true;

		Exception err = msgSend(tlsmsg, "ChangeCipher");
		if(err != null) {
			throw new RuntimeException( "scenario ChangeCipher: msgSend failed",err);
		}
		
		return;
	}
	
	public void ApplicationData() {
		if(!target.getCurrentState().connected)
			return;
		
		DebugInfo.printMessageWithSeparator(1,"scenario ApplicationData start");
		System.out.println(applicationDataString);
		ApplicationData msg = new ApplicationData(applicationDataString.getBytes());

		Exception err = msgSetContentAndSend(new RecordLayer(msg),msg,null,"ApplicationData");
		if(err != null) {
			throw new RuntimeException( "scenario ApplicationData: msgSetContentAndSend failed",err);
		}
		
		target.disconnectWithSendAlert = 0;
		target.DisconnectWithAlert(current_connection);

		return;
	}
	
	
	

	public Exception msgSetContent(Object msg, byte[] padding, String scenName) {
		return msgSetContent(null, new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContent(TLSContentType tp, Object msg, byte[] padding, String scenName) {
		return msgSetContent(tp, new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContent(RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		return msgSetContent(null, rlmsg, msg, padding, scenName);
	}
	
	public Exception msgSetContent(TLSContentType tp, RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			DebugInfo.printMessage(1," --- Object --- ");
			rlmsg.setContent(msg,padding,current_connection);
			if(tp != null) {
				rlmsg.tls_type = tp;
				rlmsg.opaque_tls_type = tp;
			}
			DebugInfo.printDump(1, rlmsg.toByteArray(), scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
//			target.disconnect();
			return res;
		}
		return res;
	}
	
	
	public Exception msgSetContent(TLSContentType tp, RecordLayer rlmsg, byte[] msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			DebugInfo.printMessage(1," --- byte[] --- ");
			rlmsg.setContent(msg,padding,current_connection);
			if(tp != null) {
				rlmsg.tls_type = tp;
				rlmsg.opaque_tls_type = tp;
			}
			DebugInfo.printDump(1, rlmsg.toByteArray(), scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
//			target.disconnect();
			return res;
		}
		return res;
	}
	
	
	public Exception msgSetContentAndSend(Object msg, byte[] padding, String scenName) {
		return msgSetContentAndSend(new RecordLayer(msg), msg, padding, scenName);
	}
	
	public Exception msgSetContentAndSend(RecordLayer rlmsg, Object msg, byte[] padding, String scenName) {
		Exception res = null;
		
		try {
			rlmsg.setContent(msg,padding,current_connection);
			DebugInfo.printDump(5, rlmsg.toByteArray(), scenName);
//			if(msg instanceof HandshakeMessage)
//				DebugInfo.printDump(5, ((HandshakeMessage)msg).net_msg, scenName);
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile, "===Error! scenario setContent Exception: " + err);
			target.disconnect();
			return res;
		}

		res = msgSend(rlmsg, scenName);
		
		return res;
	}
	
	
	public Exception msgSend(RecordLayer rlmsg, String scenName) {
		Exception res = null;
		
		try {
			ArrayList<RecordLayer> response = target.send(rlmsg,current_connection);
			if(null == response) {
				DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
				DebugInfo.logtrace(logFile,"===Error! scenario send response null, may be decrypt failed");
				target.disconnectWithSendAlert = 80;
				target.DisconnectWithAlert(current_connection);
			}
			else if(target.disconnectWithSendAlert != -1) {
				target.DisconnectWithAlert(current_connection);
			}
		}
		catch (Exception err) {
			DebugInfo.logtrace(logFile, "===Error! counter: " + testCounter + " scenario " + scenName);
			DebugInfo.logtrace(logFile,"===Error! scenario send Exception: " + err);
			target.disconnect();	
			res = err;
		}
		
		return res;
	}
	
	
	public static void main(String[] args) {
//		String fileName = "connectInfo_mail.txt";
//		String fileName = "connectInfo_http.txt";
		String fileName = "connectInfo_java.txt";
		if (args.length > 0 )
			fileName = args[0];
		start(fileName);
	}
	
	
	public static void start(String fileName) {
		String currentClassName = MethodHandles.lookup().lookupClass().getSimpleName();
		try {
			logFile = new PrintWriter(new OutputStreamWriter(new FileOutputStream(currentClassName + "_errorlog.txt",true)));
			DebugInfo.printMessage(1,"open log file " + currentClassName + "_errorlog.txt");
		}
		catch (FileNotFoundException err) {
			DebugInfo.printMessage(1,"open log file Exception: " + err);
			return;
		}
		Date dt1 = new Date();
		DebugInfo.logtrace(logFile,"start test " + currentClassName + " " + dt1.toString());
		
		LoginInfo LI = new LoginInfo();
		LI.readFromFile(fileName);
		
		TestClientHelloRetry scen = new TestClientHelloRetry(LI);

		scen.connect();
		scen.ClientHello_pre();
//		scen.ChangeCipher();
		scen.ClientHello();
		scen.Finished();
//		scen.ApplicationData();

		dt1 = new Date();
		DebugInfo.logtrace(logFile,"end test: " + dt1.toString() + " testCounter " + testCounter);
	}
}